import os
import shutil
import logging
from typing import Optional

from medcat.cat import CAT
from medcat.components.types import CoreComponentType
from medcat.storage.serialisers import AvailableSerialisers
from medcat.components.linking.no_action_linker import NoActionLinker

from medcat.utils.legacy.convert_cdb import get_cdb_from_old
from medcat.utils.legacy.convert_config import get_config_from_old
from medcat.utils.legacy.convert_vocab import get_vocab_from_old
from medcat.utils.legacy.convert_meta_cat import get_meta_cat_from_old
from medcat.utils.legacy.convert_rel_cat import get_rel_cat_from_old
from medcat.utils.legacy.convert_deid import get_trf_ner_from_old
from medcat.utils.legacy.helpers import fix_subnames


logger = logging.getLogger(__name__)


class Converter:
    """Converts v1 models to v2 models."""
    cdb_name = 'cdb.dat'
    vocab_name = 'vocab.dat'
    config_name = 'config.json'

    def __init__(self, medcat1_model_pack_path: str,
                 new_model_pack_path: Optional[str],
                 ser_type: AvailableSerialisers = AvailableSerialisers.dill):
        if medcat1_model_pack_path.endswith(".zip"):
            folder_path = medcat1_model_pack_path[:-4]
            unpack(medcat1_model_pack_path, folder_path)
            medcat1_model_pack_path = folder_path
        if not os.path.isdir(medcat1_model_pack_path):
            raise ValueError(
                "Provided model path is not a directory: "
                f"{medcat1_model_pack_path}")
        self.old_model_folder = medcat1_model_pack_path
        self.new_model_folder = new_model_pack_path
        self.ser_type = ser_type
        self._validate()

    @property
    def expected_files_in_folder(self):
        """The base names of the required files in a folder for a v1 model."""
        return [self.cdb_name, ]

    def _validate(self):
        for fn in self.expected_files_in_folder:
            path = os.path.join(self.old_model_folder, fn)
            if not os.path.exists(path):
                raise ValueError(f"Unable to find {fn} in model folder "
                                 f"{self.old_model_folder}")

    def convert(self) -> CAT:
        """Use the gathered information to convert to a v2 model.

        This converts the CDB, Vocab, and Config, in order and then
        created the model pack.

        If `self.new_model_folder` is set, the model will be saved as well.

        Returns:
            CAT: The model pack.
        """
        cdb = get_cdb_from_old(
            os.path.join(self.old_model_folder, self.cdb_name))
        vocab_path = os.path.join(self.old_model_folder, self.vocab_name)
        if os.path.exists(vocab_path):
            vocab = get_vocab_from_old(vocab_path)
        else:
            # e.g in case of DeID
            vocab = None
        cnf_path = os.path.join(self.old_model_folder, self.config_name)
        if os.path.exists(cnf_path):
            config = get_config_from_old(cnf_path)
        else:
            config = cdb.config
        cat = CAT(cdb, vocab, config)
        fix_subnames(cat)
        # MetaCATs
        meta_cats = [
            get_meta_cat_from_old(
                os.path.join(self.old_model_folder, subfolder),
                cat._pipeline.tokenizer)
            for subfolder in os.listdir(self.old_model_folder)
            if subfolder.startswith("meta_")
        ]
        for mc in meta_cats:
            cat.add_addon(mc)

        # RelCATs
        rel_cats = [
            get_rel_cat_from_old(
                cdb,
                os.path.join(self.old_model_folder, subfolder),
                cat._pipeline.tokenizer)
            for subfolder in os.listdir(self.old_model_folder)
            if subfolder.startswith("rel_")
        ]
        for rc in rel_cats:
            cat.add_addon(rc)

        # DeID / TransformersNER
        trf_ners = [
            get_trf_ner_from_old(
                os.path.join(self.old_model_folder, subfolder),
                cat._pipeline.tokenizer)
            for subfolder in os.listdir(self.old_model_folder)
            if subfolder.startswith("trf_")
        ]
        if len(trf_ners) > 1:
            raise ValueError("Cannot use more than 1 tranformers NER. "
                             f"Got {len(trf_ners)}")
        if trf_ners:
            logger.info("Found a Transformers based NER component "
                        "- probably for DeID")
            trf_ner = trf_ners[0]
            # update the TrfNER-loaded CDB with the correct config
            trf_ner._component.cdb.config = config
            # update the config with the correct NER custom config
            trf_cnf = trf_ner._component.config
            config.components.ner.custom_cnf = trf_cnf
            # update the component name
            config.components.ner.comp_name = trf_ner.name
            # replace component in pipeline
            # get the index of component in list
            index = next((c_num for c_num, comp in
                          enumerate(cat._pipeline._components)
                          if comp.get_type() is CoreComponentType.ner))
            # set / change / replace the NER component
            logger.info(f"Changing the NER component in the pipe to {trf_ner}")
            cat._pipeline._components[index] = trf_ner
            # replace linker to no-action linker
            config.components.linking.comp_name = 'no_action'
            index_link = next(
                (c_num for c_num, comp in enumerate(cat._pipeline._components)
                 if comp.get_type() is CoreComponentType.linking))
            # set / change / replace Linker to no-action linker
            logger.info("Changing the linking component in the pipe to a "
                        "no-action linker")
            cat._pipeline._components[index_link] = NoActionLinker()

        if self.new_model_folder:
            logger.info("Saving converted model to '%s'",
                        self.new_model_folder)
            cat.save_model_pack(self.new_model_folder,
                                serialiser_type=self.ser_type)
        return cat


def unpack(model_zip_path: str, target_folder: str):
    """Unpack v1 model into target folder.

    Args:
        model_zip_path (str): ZIP path.
        target_folder (str): Target folder.
    """
    shutil.unpack_archive(model_zip_path, extract_dir=target_folder)



from typing import Optional, Any, cast, Union, Literal
from typing_extensions import TypedDict
import os
import logging

# import dill
import numpy as np

from medcat.storage.serialisables import AbstractSerialisable
from medcat.storage.serialisers import (
    deserialise, AvailableSerialisers, serialise)
from medcat.storage.zip_utils import (
    should_serialise_as_zip, serialise_as_zip, deserialise_from_zip)
from medcat.utils.defaults import avoid_legacy_conversion
from medcat.utils.defaults import doing_legacy_conversion_message
from medcat.utils.defaults import LegacyConversionDisabledError


logger = logging.getLogger(__name__)


WordDescriptor = TypedDict('WordDescriptor',
                           {'vector': Optional[np.ndarray],
                            'count': int, 'index': int})


class Vocab(AbstractSerialisable):
    """Vocabulary used to store word embeddings for context similarity
    calculation. Also used by the spell checker - but not for fixing the
    spelling only for checking is something correct.

    Properties:
        vocab (dict[str, WordDescriptor]):
            Map from word to attributes, e.g. {'house':
                {'vector': <np.array>, 'count': <int>, ...}, ...}
        index2word (dict[int, str]):
            From word to an index - used for negative sampling
        vec_index2word (dict):
            Same as index2word but only words that have vectors
    """
    def __init__(self) -> None:
        super().__init__()
        self.vocab: dict[str, WordDescriptor] = {}
        self.index2word: dict[int, str] = {}
        self.vec_index2word: dict[int, str] = {}
        self.cum_probs: np.ndarray = np.array([])

    def inc_or_add(self, word: str, cnt: int = 1,
                   vec: Optional[np.ndarray] = None) -> None:
        """Add a word or increase its count.

        Args:
            word(str):
                Word to be added
            cnt(int):
                By how much should the count be increased, or to what
                should it be set if a new word. (Default value = 1)
            vec(Optional[np.ndarray]):
                Word vector (Default value = None)
        """
        if word not in self.vocab:
            self.add_word(word, cnt, vec)
        else:
            self.inc_wc(word, cnt)

    def remove_all_vectors(self) -> None:
        """Remove all stored vector representations."""
        self.vec_index2word = {}

        for word in self.vocab:
            self.vocab[word]['vector'] = None

    def remove_words_below_cnt(self, cnt: int) -> None:
        """Remove all words with frequency below cnt.

        Args:
            cnt(int):
                Word count limit.
        """
        for word in list(self.vocab.keys()):
            if self.vocab[word]['count'] < cnt:
                del self.vocab[word]

        # Rebuild index2word and vec_index2word
        self._rebuild_index()

    def _rebuild_index(self):
        self.index2word = {}
        self.vec_index2word = {}
        for word, word_info in self.vocab.items():
            ind = len(self.index2word)
            self.index2word[ind] = word
            word_info['index'] = ind

            if word_info['vector'] is not None:
                self.vec_index2word[ind] = word

    def inc_wc(self, word: str, cnt: int = 1) -> None:
        """Incraese word count by cnt.

        Args:
            word(str):
                For which word to increase the count
            cnt(int):
                By how muhc to increase the count (Default value = 1)
        """
        self.item(word)['count'] += cnt

    def add_vec(self, word: str, vec: np.ndarray) -> None:
        """Add vector to a word.

        Args:
            word(str):
                To which word to add the vector.
            vec(np.ndarray):
                The vector to add.
        """
        self.vocab[word]['vector'] = vec

        ind = self.vocab[word]['index']
        if ind not in self.vec_index2word:
            self.vec_index2word[ind] = word

    def reset_counts(self, cnt: int = 1) -> None:
        """Reset the count for all word to cnt.

        Args:
            cnt(int):
                New count for all words in the vocab. (Default value = 1)
        """
        for word in self.vocab.keys():
            self.vocab[word]['count'] = cnt

    def update_counts(self, tokens: list[str]) -> None:
        """Given a list of tokens update counts for words in the vocab.

        Args:
            tokens(list[str]):
                Usually a large block of text split into tokens/words.
        """
        for token in tokens:
            if token in self:
                self.inc_wc(token, 1)

    def add_word(self, word: str, cnt: int = 1,
                 vec: Optional[np.ndarray] = None,
                 replace: bool = True) -> None:
        """Add a word to the vocabulary

        Args:
            word (str):
                The word to be added, it should be lemmatized and lowercased
            cnt (int):
                Count of this word in your dataset (Default value = 1)
            vec (Optional[np.ndarray]):
                The vector representation of the word (Default value = None)
            replace (bool):
                Will replace old vector representation (Default value = True)
        """
        if word not in self.vocab:
            # NOTE: If one were to manually remove a word, this could have
            #       issues, but the Vocab should - in general - be pretty
            #       stable, so shouldn't be an issue
            ind = len(self.index2word)
            self.index2word[ind] = word
            item: WordDescriptor = {'vector': vec, 'count': cnt, 'index': ind}
            self.vocab[word] = item

            if vec is not None:
                self.vec_index2word[ind] = word
        elif replace and vec is not None:
            word_info = self.vocab[word]
            word_info['vector'] = vec
            word_info['count'] = cnt

            # If this word didn't have a vector before
            ind = word_info['index']
            if ind not in self.vec_index2word:
                self.vec_index2word[ind] = word

    def add_words(self, path: str, replace: bool = True) -> None:
        """Adds words to the vocab from a file, the file
        is required to have the following format (vec being optional):
            <word>\t<cnt>[\t<vec_space_separated>]

        e.g. one line: the word house with 3 dimensional vectors
            house   34444   0.3232 0.123213 1.231231

        Args:
            path(str):
                path to the file with words and vectors
            replace(bool):
                existing words in the vocabulary will be replaced.
                Defaults to True.
        """
        with open(path) as f:
            for line in f:
                parts = line.split("\t")
                word = parts[0]
                cnt = int(parts[1].strip())
                vec = None
                if len(parts) == 3:
                    floats = [float(x) for x in parts[2].strip().split(" ")]
                    vec = np.array(floats)

                self.add_word(word, cnt, vec, replace)

    def init_cumsums(self) -> None:
        """Initialise cumulative sums.

        This is in place of the unigram table. But similarly to it, this
        approach allows generating a list of indices that match the
        probabilistic distribution expected as per the word counts of each
        word.
        """
        raw_freqs = []
        index_list = []
        # index list maps the slot in which a word index
        # sits in vec_index2word to the actual index for said word
        # e.g:
        #    if we have words indexed 0, 1, and 2
        #    but only 0, and 2 have corresponding vectors
        #    then only 0 and 2 will occur in vec_index2word
        #    and while 0 will be in the 0th position (as expected)
        #    in the final probability list, 2 will be in 1st position
        #    so we need to mark that conversion down
        for word_index, word in self.vec_index2word.items():
            raw_freqs.append(self[word])
            index_list.append(word_index)

        freqs = np.array(raw_freqs) ** (3 / 4)
        freqs /= freqs.sum()

        self.cum_probs = np.cumsum(freqs)
        # the mapping from vector index order to word indices
        self._index_list = index_list

    def get_negative_samples(self, n: int = 6,
                             ignore_punct_and_num: bool = False) -> list[int]:
        """Get N negative samples.

        Args:
            n (int):
                How many words to return (Default value = 6)
            ignore_punct_and_num (bool):
                Whether to ignore punctuation and numbers. Defaults to False.

        Raises:
            Exception: If no unigram table is present.

        Returns:
            list[int]:
                Indices for words in this vocabulary.
        """
        if len(self.cum_probs) == 0:
            self.init_cumsums()
        random_vals = np.random.rand(n)
        # NOTE: These indices are in terms of the cum_probs array
        #       which only has word data for words with vectors.
        vec_slots = cast(
            list[int], np.searchsorted(self.cum_probs, random_vals).tolist())
        # so we need to translate these back to word indices
        inds = list(map(self._index_list.__getitem__, vec_slots))

        if ignore_punct_and_num:
            # Do not return anything that does not have letters in it
            return [ind for ind in inds
                    if self.index2word[ind].upper().isupper()]
        return inds

    def get_vectors(self, indices: list[int]) -> list[np.ndarray]:
        return [self.vec(self.index2word[ind])  # type: ignore
                for ind in indices if ind in self.vec_index2word]

    def __getitem__(self, word: str) -> int:
        return self.count(word)

    def vec(self, word: str) -> Optional[np.ndarray]:
        return self.vocab[word]['vector']

    def count(self, word: str) -> int:
        return self.vocab[word]['count']

    def item(self, word: str) -> WordDescriptor:
        return self.vocab[word]

    def __contains__(self, word: str) -> bool:
        if word in self.vocab:
            return True

        return False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Vocab):
            return False
        return (self.vocab.keys() == other.vocab.keys() and
                all(v1.keys() == v2.keys()
                    and all(
                        np.all(sv1 == sv2) if isinstance(sv1, np.ndarray)
                        else sv1 == sv2
                        for sv1, sv2 in zip(v1.values(), v2.values()))
                    for v1, v2
                    in zip(self.vocab.values(), other.vocab.values())) and
                self.index2word == other.index2word and
                self.vec_index2word == other.vec_index2word)

    def save(self, save_path: str,
             serialiser: Union[
                 str, AvailableSerialisers] = AvailableSerialisers.dill,
             overwrite: bool = False,
             as_zip: Union[bool, Literal['auto']] = 'auto',
             ) -> None:
        """Save Vocab at path.

        Args:
            save_path (str):
                The path to save at.
            serialiser (Union[ str, AvailableSerialisers], optional):
                The serialiser. Defaults to AvailableSerialisers.dill.
            overwrite (bool, optional):
                Whether to allow overwriting existing files. Defaults to False.
            as_zip (Union[bool, Literal['auto']]):
                Whether to serialise the CDB as a zip.
        """
        if should_serialise_as_zip(save_path, as_zip):
            serialise_as_zip(self, save_path, serialiser, overwrite=overwrite)
            return
        serialise(serialiser, self, save_path, overwrite=overwrite)

    @classmethod
    def load(cls, path: str) -> 'Vocab':
        if should_serialise_as_zip(path, 'auto'):
            vocab = deserialise_from_zip(path)
        elif os.path.isfile(path) and path.endswith('.dat'):
            if not avoid_legacy_conversion():
                from medcat.utils.legacy.convert_vocab import (
                    get_vocab_from_old)
                doing_legacy_conversion_message(logger, 'Vocab', path)
                vocab = get_vocab_from_old(path)
            else:
                raise LegacyConversionDisabledError("Vocab")
        else:
            vocab = deserialise(path)
        if not isinstance(vocab, Vocab):
            raise ValueError(f"The path '{path}' is not a Vocab!")
        return vocab



from typing import Iterable, Any, Collection, Union, Literal
import os

from medcat.storage.serialisables import AbstractSerialisable
from medcat.cdb.concepts import CUIInfo, NameInfo, TypeInfo
from medcat.cdb.concepts import get_new_cui_info, get_new_name_info
from medcat.cdb.concepts import reset_cui_training
from medcat.storage.serialisers import (
    deserialise, AvailableSerialisers, serialise)
from medcat.storage.zip_utils import (
    should_serialise_as_zip, serialise_as_zip, deserialise_from_zip)
from medcat.utils.defaults import default_weighted_average, StatusTypes as ST
from medcat.utils.defaults import avoid_legacy_conversion
from medcat.utils.defaults import doing_legacy_conversion_message
from medcat.utils.defaults import LegacyConversionDisabledError
from medcat.utils.hasher import Hasher
from medcat.preprocessors.cleaners import NameDescriptor
from medcat.config import Config
from medcat.data.model_card import CDBInfo

import logging


logger = logging.getLogger(__name__)


class CDB(AbstractSerialisable):

    def __init__(self, config: Config) -> None:
        self.config = config
        self.cui2info: dict[str, CUIInfo] = {}
        self.name2info: dict[str, NameInfo] = {}
        self.type_id2info: dict[str, TypeInfo] = {}
        self.token_counts: dict[str, int] = {}
        self.addl_info: dict[str, Any] = {}
        self._subnames: set[str] = set()
        self.is_dirty = False
        self.has_changed_names = False

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return ['config']

    def _reset_subnames(self):
        logger.info("Resetting subnames")
        self._subnames.clear()
        for info in self.cui2info.values():
            self._subnames.update(info['subnames'])
        self.has_changed_names = False

    def has_subname(self, name: str) -> bool:
        """Whether the CDB has the specified subname.

        Args:
            name (str): The subname to check.

        Returns:
            bool: Whether the subname is present in this CDB.
        """
        if (self.has_changed_names or
                len(self._subnames) < len(self.name2info)):
            self._reset_subnames()
        return name in self._subnames

    def get_name(self, cui: str) -> str:
        """Returns preferred name if it exists, otherwise it will return
        the longest name assigned to the concept.

        Args:
            cui (str):
                Concept ID or unique identifier in this database.

        Returns:
            str: The name of the concept.
        """
        name = cui
        sep = self.config.general.separator

        if cui not in self.cui2info:
            return name
        cui_info = self.cui2info[cui]
        pref_name = cui_info['preferred_name']
        names = cui_info['names']
        if pref_name:
            name = pref_name
        elif names:
            # longest name
            raw_name = max(names, key=len)
            name = " ".join(str(raw_name).split(sep)).title()

        return name

    def weighted_average_function(self, step: int) -> float:
        """Get the weighted average for steop.

        Args:
            step (int): The steop.

        Returns:
            float: The weighted average.
        """
        return default_weighted_average(step)

    def add_types(self, types: Iterable[tuple[str, str]]) -> None:
        """Add type info to CDB.

        Args:
            types (Iterable[tuple[str, str]]): The raw type info.
        """
        for type_id, name in types:
            self.type_id2info[type_id] = TypeInfo(type_id, name)

    def add_names(self, cui: str, names: dict[str, NameDescriptor],
                  name_status: str = ST.AUTOMATIC, full_build: bool = False
                  ) -> None:
        """Adds a name to an existing concept.

        Args:
            cui (str):
                Concept ID or unique identifier in this database, all concepts
                that have the same CUI will be merged internally.
            names (dict[str, NameDescriptor]):
                Names for this concept, or the value that if found in free
                text can be linked to this concept. Names is an dict like:
                `{name: {'tokens': tokens, 'snames': snames,
                         'raw_name': raw_name}, ...}`
                Names should be generated by helper function
                'medcat.preprocessing.cleaners.prepare_name'
            name_status (str):
                One of `P`, `N`, `A`. Defaults to 'A'.
            full_build (bool):
                If True the dictionary self.addl_info will also be populated,
                contains a lot of extra information about concepts, but can be
                very memory consuming. This is not necessary for normal
                functioning of MedCAT (Default value `False`).
        """
        name_status = name_status.upper()
        if name_status not in ST.ALLOWED_STATUS:
            # Name status must be one of the three
            name_status = ST.AUTOMATIC

        self._add_concept(cui=cui, names=names, ontologies=set(),
                          name_status=name_status, type_ids=set(),
                          description='', full_build=full_build)

    def _add_concept_names(self, cui: str, names: dict[str, NameDescriptor],
                           name_status: str) -> None:
        cui_info = self.cui2info[cui]
        for name, in_name_info in names.items():
            # add name and synonyms
            cui_info['names'].add(name)
            cui_info['subnames'].update(in_name_info.snames)

            if name not in self.name2info:
                self.name2info[name] = get_new_name_info(name=name)
            # Add whether concept is uppercase
            name_info = self.name2info[name]
            name_info['is_upper'] = in_name_info.is_upper
            status_map = name_info['per_cui_status']
            if cui not in status_map:
                status_map[cui] = name_status
            elif name_status == ST.PRIMARY_STATUS_NO_DISAMB:
                # if this is primary, overwrite old status
                status_map[cui] = name_status
            # if already in status map and other status, leave it be

            # Add tokens to token counts
            for token in in_name_info.tokens:
                if token in self.token_counts:
                    self.token_counts[token] += 1
                else:
                    self.token_counts[token] = 1
            self._subnames.update(cui_info['subnames'])
            self.is_dirty = True

    def _add_full_build(self, cui: str, names: dict[str, NameDescriptor],
                        ontologies: set[str], description: str,
                        type_ids: set[str]) -> None:
        cui_info = self.cui2info[cui]
        # Use original_names as the base check because they must be added
        orig_names: set[str] = set([v.raw_name for v in names.values()])
        if cui_info['original_names'] is None:
            if ontologies:
                cui_info['in_other_ontology'] = ontologies
            cui_info['original_names'] = orig_names
        else:
            # Update existing ones
            if ontologies:
                ontos = cui_info['in_other_ontology']
                if ontos is None:
                    ontos = cui_info['in_other_ontology'] = set()
                ontos.update(ontologies)
            cui_info['original_names'].update(orig_names)
        if description:
            cui_info['description'] = description

        for type_id in type_ids:
            # Add type_id2cuis link
            if type_id not in self.type_id2info:
                # NOTE: most type IDs should be added at CDB creation and
                #       their names should be (manually?) added alongside
                self.type_id2info[type_id] = TypeInfo(type_id=type_id,
                                                      name='N/A')
            type_info = self.type_id2info[type_id]
            type_info.cuis.add(cui)

    def _add_concept(self,
                     cui: str,
                     names: dict[str, NameDescriptor],
                     ontologies: set[str],
                     name_status: str,
                     type_ids: set[str],
                     description: str,
                     full_build: bool = False) -> None:
        """Add a concept to internal Concept Database (CDB). Depending on what
        you are providing this will add a large number of properties for each
        concept.

        Args:
            cui (str):
                Concept ID or unique identifier in this database, all concepts
                that have the same CUI will be merged internally.
            names (dict[str, NameDescriptor]):
                Names for this concept, or the value that if found in free
                text can be linked to this concept. Names is a dict like:
                `{name: {'tokens': tokens, 'snames': snames,
                         'raw_name': raw_name}, ...}`
                Names should be generated by helper function
                'medcat.preprocessing.cleaners.prepare_name'
            ontologies (set[str]):
                ontologies in which the concept exists (e.g. SNOMEDCT, HPO)
            name_status (str):
                One of `P`, `N`, `A`
            type_ids (set[str]):
                Semantic type identifier (have a look at TUIs in UMLS or
                SNOMED-CT)
            description (str):
                Description of this concept.
            full_build (bool):
                If True the dictionary self.addl_info will also be populated,
                contains a lot of extra information about concepts, but can be
                very memory consuming. This is not necessary for normal
                functioning of MedCAT (Default Value `False`).
        """
        if not names:
            logger.warning("Passed an empty names dict for CUI '%s'. "
                           "This could be caused by the preprocessor "
                           "not picking up names due to the value of "
                           "'config.cdb_maker.min_letters_required' "
                           "(currently %d) being too small for this "
                           "particular name", cui,
                           self.config.cdb_maker.min_letters_required)
            return
        # Add CUI to the required dictionaries
        if cui not in self.cui2info:
            # Create placeholders
            cui_info = get_new_cui_info(
                cui=cui, preferred_name='', type_ids=type_ids)
            self.cui2info[cui] = cui_info
        else:
            cui_info = self.cui2info[cui]
            # If the CUI is already in update the type_ids
            cui_info['type_ids'].update(type_ids)

        # Add names to the required dictionaries
        self._add_concept_names(cui, names, name_status)

        if name_status == 'P' and not cui_info['preferred_name']:
            raw_names = [ini.raw_name for ini in names.values()]
            # TODO: which raw name?
            # previous implementation used the somewhat arbitrary last raw name
            rni = -1
            cui_info['preferred_name'] = raw_names[rni]

        # Add other fields if full_build
        if full_build:
            self._add_full_build(cui, names, ontologies, description, type_ids)
        self.is_dirty = True

    def reset_training(self) -> None:
        """Will remove all training efforts - in other words all embeddings
        that are learnt for concepts in the current CDB. Please note that this
        does not remove synonyms (names) that were potentially added during
        supervised/online learning.
        """
        for cui_info in self.cui2info.values():
            reset_cui_training(cui_info)
        for name_info in self.name2info.values():
            name_info['count_train'] = 0
        self._subnames.clear()
        # clear config entries as well
        self.config.meta.unsup_trained.clear()
        self.config.meta.sup_trained.clear()
        self.is_dirty = True

    def filter_by_cui(self, cuis_to_keep: Collection[str]) -> None:
        """Subset the core CDB fields (dictionaries/maps).

        Note that this will potenitally keep a bit more CUIs
        then in cuis_to_keep. It will first find all names that
        link to the cuis_to_keep and then find all CUIs that
        link to those names and keep all of them.

        This also will not remove any data from cdb.addl_info -
        as this field can contain data of unknown structure.

        Args:
            cuis_to_keep (Collection[str]):
                CUIs that will be kept, the rest will be removed
                (not completely, look above).

        Raises:
            Exception: If no snames and subsetting is not possible.
        """
        # First get all names/snames that should be kept based on this CUIs
        names_to_keep = set()
        snames_to_keep = set()
        for cui in cuis_to_keep:
            if cui not in self.cui2info:
                logger.warning(
                    "While filtering for CUIs asked to keep CUI '%s'"
                    "which is not a part of the existing CDB", cui)
                continue
            ci = self.cui2info[cui]
            names_to_keep.update(ci['names'])
            snames_to_keep.update(ci['subnames'])

        # Based on the names get also the indirect CUIs that have to be kept
        all_cuis_to_keep: set[str] = set()
        for name in names_to_keep:
            # NOTE: since this was based on the cui2info they
            #       should all have a name info
            ni = self.name2info[name]
            all_cuis_to_keep.update(ni['per_cui_status'].keys())

        new_cui2info: dict[str, CUIInfo] = {}
        new_name2info: dict[str, NameInfo] = {}

        # get kept
        for cui in all_cuis_to_keep:
            if cui not in self.cui2info:
                # NOTE: already warned above
                continue
            new_cui2info[cui] = self.cui2info[cui]

        for name in names_to_keep:
            # NOTE: should all be in name2info since got from cui2info
            new_name2info[name] = self.name2info[name]

        # set filtered dicts
        self.cui2info = new_cui2info
        self.name2info = new_name2info
        # redo all subnames
        self._reset_subnames()
        self.is_dirty = True

    def remove_cui(self, cui: str) -> None:
        """This function takes a CUI and removes it the CDB.

        It also removes the CUI from name specific per_cui_status
        maps as well as well as removes all the names that do not
        correspond to any CUIs after the removal of this one.

        Args:
            cui (str): The CUI to remove.
        """
        if cui not in self.cui2info:
            logger.warning(
                "Trying remove CUI '%s' which does not exist in CDB", cui)
            return
        ci = self.cui2info.pop(cui)
        for name in ci['names']:
            ni = self.name2info[name]
            del ni['per_cui_status'][cui]
            # if name name corresponds to no CUIs
            if not ni['per_cui_status']:
                del self.name2info[name]
        # need to reset subnames
        self._reset_subnames()
        self.is_dirty = True

    def _remove_names(self, cui: str, names: Iterable[str]) -> None:
        """Remove names from an existing concept - effect is this name will
        never again be used to link to this concept. This will only remove the
        name from the linker (namely name2cuis and name2cuis2status), the name
        will still be present everywhere else. Why? Because it is bothersome
        to remove it from everywhere, but could also be useful to keep the
        removed names in e.g. cui2names.

        Args:
            cui (str):
                Concept ID or unique identifier in this database.
            names (Iterable[str]):
                Names to be removed (e.g list, set, or even a dict (in which
                case keys will be used)).
        """
        for name in names:
            if name in self.name2info:
                info = self.name2info[name]
                if cui in info['per_cui_status']:
                    del info['per_cui_status'][cui]
                if len(info['per_cui_status']) == 0:
                    del self.name2info[name]

            # Remove from per_cui_status
            if name in self.name2info:
                info = self.name2info[name]
                cuis2status = info['per_cui_status']
                if cui in cuis2status:
                    _ = cuis2status.pop(cui)
                if len(cuis2status) == 0:
                    # TODO: does this make sense?
                    del self.name2info[name]

            # Set to disamb always if name2cuis2status is now only one CUI
            if name in self.name2info:
                info = self.name2info[name]
                cuis2status = info['per_cui_status']
                if len(cuis2status) == 1:
                    for _cui in cuis2status:
                        if cuis2status[_cui] == 'A':
                            cuis2status[_cui] = 'N'
                        elif cuis2status[_cui] == 'P':
                            cuis2status[_cui] = 'PD'
        self.is_dirty = True
        self.has_changed_names = True

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, CDB):
            return False
        # NOTE: Using config.model_dump since
        #       some parts of the config should not be considered.
        #       This refers to (mostly) the init args stored within there
        #       for various components
        return (self.config.model_dump() == other.config.model_dump() and
                self.cui2info == other.cui2info and
                self.name2info == other.name2info and
                self.type_id2info == other.type_id2info and
                self.token_counts == other.token_counts)

    def get_cui2count_train(self) -> dict[str, int]:
        return {
            cui: ct for cui, ci in self.cui2info.items()
            if (ct := ci['count_train'])
        }

    def get_name2count_train(self) -> dict[str, int]:
        return {
            cui: ct for cui, ni in self.name2info.items()
            if (ct := ni['count_train'])
        }

    def get_hash(self) -> str:
        hasher = Hasher()
        # only length for number of cuis/names/subnames
        hasher.update(len(self.cui2info))
        hasher.update(len(self.name2info))
        hasher.update(len(self._subnames))
        # the entirety of trained stuff
        hasher.update(self.get_cui2count_train())
        hasher.update(self.get_name2count_train())
        return hasher.hexdigest()

    def get_basic_info(self) -> CDBInfo:
        cui2ct = self.get_cui2count_train()
        cuis_trained = len(cui2ct)
        examples_seen = sum(cui2ct.values())
        if cuis_trained:
            average_count_train = examples_seen / cuis_trained
        else:
            average_count_train = 0.5
        unsup_history = [
            history.model_dump(mode='json')
            for history in self.config.meta.unsup_trained
        ]
        sup_history = [
            history.model_dump(mode='json')
            for history in self.config.meta.sup_trained
        ]
        return {
            "Number of concepts": len(self.cui2info),
            "Number of names": len(self.name2info),
            "Number of concepts that received training": cuis_trained,
            "Number of seen training examples in total": examples_seen,
            "Average training examples per concept": average_count_train,
            "Unsupervised training history": unsup_history,
            "Supervised training history": sup_history,
        }

    def save(self, save_path: str,
             serialiser: Union[
                 str, AvailableSerialisers] = AvailableSerialisers.dill,
             overwrite: bool = False,
             as_zip: Union[bool, Literal['auto']] = 'auto',
             ) -> None:
        """Save CDB at path.

        Args:
            save_path (str):
                The path to save at.
            serialiser (Union[ str, AvailableSerialisers], optional):
                The serialiser. Defaults to AvailableSerialisers.dill.
            overwrite (bool, optional):
                Whether to allow overwriting existing files. Defaults to False.
            as_zip (Union[bool, Literal['auto']]):
                Whether to serialise the CDB as a zip.
        """
        if should_serialise_as_zip(save_path, as_zip):
            serialise_as_zip(self, save_path, serialiser, overwrite=overwrite)
            return
        serialise(serialiser, self, save_path, overwrite=overwrite)

    @classmethod
    def load(cls, path: str) -> 'CDB':
        if should_serialise_as_zip(path, 'auto'):
            cdb = deserialise_from_zip(path)
        elif os.path.isfile(path) and path.endswith('.dat'):
            if not avoid_legacy_conversion():
                from medcat.utils.legacy.convert_cdb import get_cdb_from_old
                doing_legacy_conversion_message(logger, 'CDB', path)
                cdb = get_cdb_from_old(path)
            else:
                raise LegacyConversionDisabledError("CDB")
        else:
            cdb = deserialise(path)
        if not isinstance(cdb, CDB):
            raise ValueError(f"The path '{path}' is not a CDB!")
        return cdb



import os
import json
import re
import hashlib
import pandas as pd
from typing import Dict, List, Optional, Tuple
from dataclasses import dataclass, field
from enum import Enum, auto


def parse_file(filename, first_row_header=True, columns=None):
    with open(filename, encoding='utf-8') as f:
        entities = [[n.strip() for n in line.split('\t')] for line in f]
        return pd.DataFrame(
            entities[1:], columns=entities[0] if first_row_header else columns)


def get_all_children(sctid, pt2ch):
    """
    Retrieves all the children of a given SNOMED CT ID (SCTID) from a given
    parent-to-child mapping (pt2ch) via the "IS A" relationship.
    pt2ch can be found in a MedCAT model in the additional info
    via the call: cat.cdb.addl_info['pt2ch']

    Args:
        sctid (int): The SCTID whose children need to be retrieved.
        pt2ch (dict): A dictionary containing the parent-to-child
            elationships in the form {parent_sctid: [list of child sctids]}.

    Returns:
        list: A list of unique SCTIDs that are children of the given SCTID.
    """
    result = []
    stack = [sctid]
    while len(stack) != 0:
        # remove the last element from the stack
        current_snomed = stack.pop()
        current_snomed_children = pt2ch.get(current_snomed, [])
        stack.extend(current_snomed_children)
        result.append(current_snomed)
    result = list(set(result))
    return result


def get_direct_refset_mapping(in_dict: dict) -> dict:
    """This method uses the output from Snomed.map_snomed2icd10 or
    Snomed.map_snomed2opcs4 and removes the metadata and maps each
    SNOMED CUI to the prioritised list of the target ontology CUIs.

    The input dict is expected to be in the following format:
    - Keys are SnomedCT CUIs
    - The values are lists of dictionaries, each list item (at least)
      - Has a key 'code' that specifies the target onotlogy CUI
      - Has a key 'mapPriority' that specifies the priority

    Args:
        in_dict (dict): The input dict.

    Returns:
        dict: The map from Snomed CUI to list of priorities list of target
            ontology CUIs.
    """
    ret_dict = dict()
    for k, vals in in_dict.items():
        # sort such that highest priority values are first
        svals = sorted(vals, key=lambda el: el['mapPriority'], reverse=True)
        # only keep the code / CUI
        ret_dict[k] = [v['code'] for v in svals]
    return ret_dict


_IGNORE_TAG = '##IGNORE-THIS##'


class RefSetFileType(Enum):
    concept = auto()
    description = auto()
    relationship = auto()
    refset = auto()


@dataclass
class FileFormatDescriptor:
    concept: str
    description: str
    relationship: str
    refset: str
    # for concept, description, and relationship (but not refset)
    common_prefix: str = "sct2_"

    @classmethod
    def ignore_all(cls) -> 'FileFormatDescriptor':
        return cls(concept=_IGNORE_TAG, description=_IGNORE_TAG,
                   relationship=_IGNORE_TAG, refset=_IGNORE_TAG)

    def get_file_per_type(self, file_type: RefSetFileType) -> str:
        raw = self._get_raw(file_type)
        return (raw
                if file_type == RefSetFileType.refset else
                self.common_prefix + raw)

    def _get_raw(self, file_type: RefSetFileType) -> str:
        return getattr(self, file_type.name)

    def get_concept(self) -> str:
        return self.get_file_per_type(RefSetFileType.concept)

    def get_description(self) -> str:
        return self.get_file_per_type(RefSetFileType.description)

    def get_relationship(self) -> str:
        return self.get_file_per_type(RefSetFileType.relationship)

    def get_refset(self) -> str:
        return self.get_file_per_type(RefSetFileType.refset)


@dataclass
class ExtensionDescription:
    exp_name_in_folder: str
    exp_files: FileFormatDescriptor
    exp_2nd_part_in_folder: Optional[str] = None


SNOMED_FOLDER_NAME_PATTERN = re.compile(
    # within:     EXTENSION      PRODUCTION     RELEASE
    r"^SnomedCT_([A-Za-z0-9]+)_([A-Za-z0-9]+)_(\d{8}T\d{6}Z$)")
PER_FILE_TYPE_PATHS = {
    RefSetFileType.concept: os.path.join("Snapshot", "Terminology"),
    RefSetFileType.description: os.path.join("Snapshot", "Terminology"),
    RefSetFileType.relationship: os.path.join("Snapshot", "Terminology"),
    RefSetFileType.refset: os.path.join("Snapshot", "Refset", "Map"),
}


class SupportedExtension(Enum):
    INTERNATIONAL = ExtensionDescription(
        exp_name_in_folder="InternationalRF2",
        exp_files=FileFormatDescriptor(
            concept="Concept_Snapshot",
            description="Description_Snapshot-en",
            relationship="Relationship_Snapshot",
            # NOTE: the below will be ignored for UK_CLIN bundle
            refset="der2_iisssccRefset_ExtendedMapSnapshot"
        ),
    )
    UK_CLINICAL = ExtensionDescription(
        exp_name_in_folder="UKClinicalRF2",
        exp_files=FileFormatDescriptor(
            concept="Concept_UKCLSnapshot",
            description="Description_UKCLSnapshot-en",
            relationship="Relationship_UKCLSnapshot",
            refset="der2_iisssciRefset_ExtendedMapUKCLSnapshot"
        ),
    )
    UK_CLINICAL_REFSET = ExtensionDescription(
        exp_name_in_folder="UKClinicalRefsetsRF2",
        exp_files=FileFormatDescriptor.ignore_all()
    )
    UK_EDITION = ExtensionDescription(
        exp_name_in_folder="UKEditionRF2",
        exp_files=FileFormatDescriptor(
            concept="Concept_UKEDSnapshot",
            description="Description_UKEDSnapshot-en",
            relationship="Relationship_UKEDSnapshot",
            refset="der2_iisssciRefset_ExtendedMapUKEDSnapshot"
        ),
    )
    UK_DRUG = ExtensionDescription(
        exp_name_in_folder="UKDrugRF2",
        exp_files=FileFormatDescriptor(
            concept="Concept_UKDGSnapshot",
            description="Description_UKDGSnapshot-en",
            relationship="Relationship_UKDGSnapshot",
            refset="der2_iisssciRefset_ExtendedMapUKDGSnapshot",
        ),
    )
    AU = ExtensionDescription(
        exp_name_in_folder="Release",
        exp_2nd_part_in_folder="AU1000036",
        exp_files=FileFormatDescriptor(
            concept="Concept_Snapshot",
            description="Description_Snapshot-en-AU",
            relationship="Relationship_Snapshot",
            refset=_IGNORE_TAG,
        ),
    )


@dataclass
class BundleDescriptor:
    extensions: List[SupportedExtension]
    ignores: Dict[RefSetFileType, List[SupportedExtension]] = field(
        default_factory=dict)

    def has_invalid(self, ext: SupportedExtension,
                    file_types: Tuple[RefSetFileType]) -> bool:
        for ft in file_types:
            if ft not in self.ignores:
                continue
            exts2ignore = self.ignores[ft]
            if ext in exts2ignore:
                return True
        return False


class SupportedBundles(Enum):
    UK_CLIN = BundleDescriptor(
        extensions=[
            SupportedExtension.INTERNATIONAL, SupportedExtension.UK_CLINICAL,
            SupportedExtension.UK_CLINICAL_REFSET,
            SupportedExtension.UK_EDITION],
        ignores={RefSetFileType.refset: [SupportedExtension.INTERNATIONAL]}
        )
    UK_DRUG_EXT = BundleDescriptor(
        extensions=[SupportedExtension.UK_DRUG, SupportedExtension.UK_EDITION],
        )


def match_partials_with_folders(exp_names: List[Tuple[str, Optional[str]]],
                                folder_names: List[str],
                                _group_nr1: int = 1, _group_nr2: int = 2
                                ) -> bool:
    if len(exp_names) > len(folder_names):
        return False
    available_folders = [os.path.basename(f) for f in folder_names]
    for exp_name, exp_name_p2 in exp_names:
        found_cur_name = False
        for fi, folder in enumerate(available_folders):
            m = SNOMED_FOLDER_NAME_PATTERN.match(folder)
            if not m:
                continue
            if m.group(_group_nr1) != exp_name:
                continue
            if exp_name_p2 and m.group(_group_nr2) != exp_name_p2:
                continue
            found_cur_name = True
            break
        if found_cur_name:
            available_folders.pop(fi)
        else:
            return False
    return True


class Snomed:
    """
    Pre-process SNOMED CT release files.

    This class is used to create a SNOMED CT concept DataFrame ready for
    MedCAT CDB creation.

    Attributes:
        data_path (str): Path to the unzipped SNOMED CT folder.
        release (str): Release of SNOMED CT folder.
        uk_ext (bool, optional): Specifies whether the version is a
            SNOMED UK extension released after 2021. Defaults to False.
        uk_drug_ext (bool, optional): Specifies whether the version
            is a SNOMED UK drug extension. Defaults to False.
        au_ext (bool, optional): Specifies whether the version is a
            AU release. Defaults to False.
    """
    NO_VERSION_DETECTED = 'N/A'

    def __init__(self, data_path):
        self.data_path = data_path
        self.bundle = self._determine_bundle(self.data_path)
        self.paths, self.snomed_releases, self.exts = (
            self._check_path_and_release())

    @classmethod
    def _determine_bundle(cls, data_path) -> Optional[SupportedBundles]:
        if not os.path.exists(data_path) or not os.path.isdir(data_path):
            return None
        for bundle in SupportedBundles:
            folder_names = list(os.listdir(data_path))
            exp_names = [(ext.value.exp_name_in_folder,
                          ext.value.exp_2nd_part_in_folder)
                         for ext in bundle.value.extensions]
            if match_partials_with_folders(exp_names, folder_names):
                return bundle
        return None

    def _set_extension(self, release: str, extension: SupportedExtension
                       ) -> None:
        # NOTE: now using the later refset IF by default
        # NOTE: the OPCS4 refset ID is only relevant for UK releases
        self.opcs_refset_id = '1382401000000109'
        if (extension in (SupportedExtension.UK_CLINICAL,
                          SupportedExtension.UK_DRUG) and
                # using lexicographical comparison below
                # e.g "20240101" > "20231122" results in True
                # yet "20231121" > "20231122" results in False
                len(release) == len("20231122") and release < "20231122"):
            # NOTE for UK extensions starting from 20231122 the
            #      OPCS4 refset ID seems to be different
            self.opcs_refset_id = "1126441000000105"
        self._extension = extension

    @classmethod
    def _determine_extension(cls, folder_path: str,
                             _group_nr1: int = 1, _group_nr2: int = 2
                             ) -> SupportedExtension:
        folder_basename = os.path.basename(folder_path)
        m = SNOMED_FOLDER_NAME_PATTERN.match(folder_basename)
        if not m:
            raise UnkownSnomedReleaseException(
                f"Unable to determine extension for path {repr(folder_path)}. "
                f"Checking against pattern {SNOMED_FOLDER_NAME_PATTERN}")
        ext_str = m.group(_group_nr1)
        ext_str2 = m.group(_group_nr2)
        for extension in SupportedExtension:
            if extension.value.exp_name_in_folder != ext_str:
                continue
            if (extension.value.exp_2nd_part_in_folder and
                    extension.value.exp_2nd_part_in_folder != ext_str2):
                continue
            return extension
        ext_names_folders = ",".join(
            [f"{ext.name} ({ext.value.exp_name_in_folder})"
             for ext in SupportedExtension])
        raise UnkownSnomedReleaseException(
            f"Cannot Find the extension for {folder_path}. "
            f"Tried the following extensions: {ext_names_folders}")

    @classmethod
    def _determine_release(cls, folder_path: str, strict: bool = True,
                           _group_nr: int = 3, _keep_chars: int = 8) -> str:
        folder_basename = os.path.basename(folder_path)
        match = SNOMED_FOLDER_NAME_PATTERN.match(folder_basename)
        if match is None and strict:
            raise UnkownSnomedReleaseException(
                f"No version found in '{folder_path}'")
        elif match is None:
            return cls.NO_VERSION_DETECTED
        return match.group(_group_nr)[:_keep_chars]

    def to_concept_df(self):
        """
        Create a SNOMED CT concept DataFrame.

        Creates a SNOMED CT concept DataFrame ready for MEDCAT CDB creation.
        Checks if the version is a UK extension release and sets the correct
        file names for the concept and description snapshots accordingly.
        Additionally, handles the divergent release format of the UK Drug
        Extension >v2021 with the `uk_drug_ext` variable.

        Returns:
            pandas.DataFrame: SNOMED CT concept DataFrame.
        """

        df2merge = []
        for i, snomed_release in enumerate(self.snomed_releases):
            self._set_extension(snomed_release, self.exts[i])
            contents_path = os.path.join(
                self.paths[i], PER_FILE_TYPE_PATHS[RefSetFileType.concept])
            exp_files = self._extension.value.exp_files
            concept_snapshot = exp_files.get_concept()
            description_snapshot = exp_files.get_description()
            if concept_snapshot is None or _IGNORE_TAG in concept_snapshot or (
                    self.bundle and self.bundle.value.has_invalid(
                        self._extension, [RefSetFileType.concept,
                                          RefSetFileType.description])):
                continue

            for f in os.listdir(contents_path):
                m = re.search(f'{concept_snapshot}' + r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)

            int_terms = parse_file(
                f'{contents_path}/'
                f'{concept_snapshot}_{snomed_v}_{snomed_release}.txt')
            active_terms = int_terms[int_terms.active == '1']
            del int_terms

            int_desc = parse_file(
                f'{contents_path}/{description_snapshot}_{snomed_v}_'
                f'{snomed_release}.txt')
            active_descs = int_desc[int_desc.active == '1']
            del int_desc

            _ = pd.merge(active_terms, active_descs,
                         left_on=['id'], right_on=['conceptId'],
                         how='inner')
            del active_terms
            del active_descs

            active_with_primary_desc = _[
                _['typeId'] == '900000000000003001']  # active description
            active_with_synonym_desc = _[
                _['typeId'] == '900000000000013009']  # active synonym
            del _
            active_with_all_desc = pd.concat(
                [active_with_primary_desc, active_with_synonym_desc])

            active_snomed_df = active_with_all_desc[['id_x', 'term', 'typeId']]
            del active_with_all_desc

            active_snomed_df = active_snomed_df.rename(
                columns={'id_x': 'cui', 'term': 'name',
                         'typeId': 'name_status'})
            active_snomed_df['ontologies'] = 'SNOMED-CT'
            active_snomed_df['name_status'] = active_snomed_df[
                'name_status'].replace(
                    ['900000000000003001', '900000000000013009'],
                    ['P', 'A'])
            active_snomed_df = active_snomed_df.reset_index(drop=True)

            temp_df = active_snomed_df[
                active_snomed_df['name_status'] == 'P'][['cui', 'name']]
            temp_df['description_type_ids'] = temp_df['name'].str.extract(
                r"\((\w+\s?.?\s?\w+.?\w+.?\w+.?)\)$")
            active_snomed_df = pd.merge(
                active_snomed_df,
                temp_df.loc[:, ['cui', 'description_type_ids']],
                on='cui',
                how='left')
            del temp_df

            # Hash semantic tag to get a 8 digit type_id code
            active_snomed_df['type_ids'] = (
                active_snomed_df['description_type_ids'].apply(
                    lambda x: int(
                        hashlib.sha256(str(x).encode('utf-8')).hexdigest(),
                        16) % 10 ** 8))
            df2merge.append(active_snomed_df)

        return pd.concat(df2merge).reset_index(drop=True)

    def list_all_relationships(self):
        """
        List all SNOMED CT relationships.

        SNOMED CT provides a rich set of inter-relationships between concepts.

        Returns:
            list: List of all SNOMED CT relationships.
        """
        all_rela = []
        for i, snomed_release in enumerate(self.snomed_releases):
            self._set_extension(snomed_release, self.exts[i])
            contents_path = os.path.join(
                self.paths[i], PER_FILE_TYPE_PATHS[RefSetFileType.concept])
            exp_files = self._extension.value.exp_files
            concept_snapshot = exp_files.get_concept()
            relationship_snapshot = exp_files.get_relationship()
            if concept_snapshot is None or _IGNORE_TAG in concept_snapshot or (
                    self.bundle and self.bundle.value.has_invalid(
                        self._extension, [RefSetFileType.concept,
                                          RefSetFileType.description])):
                continue

            for f in os.listdir(contents_path):
                m = re.search(f'{concept_snapshot}' + r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)
            int_relat = parse_file(
                f'{contents_path}/'
                f'{relationship_snapshot}_{snomed_v}_{snomed_release}.txt')
            active_relat = int_relat[int_relat.active == '1']
            del int_relat

            all_rela.extend(
                [relationship for
                 relationship in active_relat["typeId"].unique()])
        return all_rela

    def relationship2json(self, relationshipcode, output_jsonfile):
        """
        Convert a single relationship map structure to JSON file.

        Args:
            relationshipcode (str): A single SCTID or unique concept identifier
                of the relationship type.
            output_jsonfile (str): Name of JSON file output.

        Returns:
            file: JSON file of relationship mapping.
        """
        output_dict = {}
        for i, snomed_release in enumerate(self.snomed_releases):
            self._set_extension(snomed_release, self.exts[i])
            contents_path = os.path.join(
                self.paths[i], PER_FILE_TYPE_PATHS[RefSetFileType.concept])
            exp_files = self._extension.value.exp_files
            concept_snapshot = exp_files.get_concept()
            relationship_snapshot = exp_files.get_relationship()
            if concept_snapshot is None or _IGNORE_TAG in concept_snapshot or (
                    self.bundle and self.bundle.value.has_invalid(
                        self._extension, [RefSetFileType.concept,
                                          RefSetFileType.description])):
                continue

            for f in os.listdir(contents_path):
                m = re.search(f'{concept_snapshot}' + r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)
            int_relat = parse_file(
                f'{contents_path}/'
                f'{relationship_snapshot}_{snomed_v}_{snomed_release}.txt')
            active_relat = int_relat[int_relat.active == '1']
            del int_relat

            relationship = dict(
                [(key, []) for key in active_relat["destinationId"].unique()])
            for _, v in active_relat.iterrows():
                if v['typeId'] == str(relationshipcode):
                    _ = v['destinationId']
                    relationship[_].append(v['sourceId'])
                else:
                    pass
            output_dict = {
                key: output_dict.get(key, []) + relationship.get(key, [])
                for key in
                set(list(output_dict.keys()) + list(relationship.keys()))}
        with open(output_jsonfile, 'w') as json_file:
            json.dump(output_dict, json_file)
        return

    def map_snomed2icd10(self):
        """
        This function maps SNOMED CT concepts to ICD-10 codes using the refset
        mappings provided in the SNOMED CT release package.

        Returns:
            dict: A dictionary containing the SNOMED CT to ICD-10 mappings
                including metadata.
        """
        snomed2icd10df = self._map_snomed2refset()
        return self._refset_df2dict(snomed2icd10df[0])

    def map_snomed2opcs4(self) -> dict:
        """
        This function maps SNOMED CT concepts to OPCS-4 codes using the refset
        mappings provided in the SNOMED CT release package.

        Then it calls the internal function _map_snomed2refset() to get the
            DataFrame containing the OPCS-4 mappings.
        The function then converts the DataFrame to a dictionary using the
            internal function _refset_df2dict()

        Raises:
            AttributeError: If OPCS-4 mappings aren't available.

        Returns:
            dict: A dictionary containing the SNOMED CT to OPCS-4 mappings
                including metadata.
        """
        if all(ext not in (SupportedExtension.UK_CLINICAL,
                           SupportedExtension.UK_DRUG)
               for ext in self.exts):
            raise AttributeError(
                "OPCS-4 mapping does not exist in this edition")
        snomed2opcs4df = self._map_snomed2refset()[1]
        return self._refset_df2dict(snomed2opcs4df)

    def _check_path_and_release(self):
        """
        This function checks the path and release of the SNOMED CT data
        provided.

        It looks for the "Snapshot" folder within the data path, and if
        it's not found, it looks for any folder containing the name "SnomedCT".
        It then stores the path and release in separate lists.
        If no valid paths are found, it raises a FileNotFoundError.

        Returns:
            tuple: a tuple containing two lists, the first one is a list of
                the paths where the data is located and the second is a list
                of the releases of the data.

        Raises:
            FileNotFoundError: If the path to the SNOMED CT directory is
                incorrect.
        """
        snomed_releases = []
        paths = []
        exts = []
        if "Snapshot" in os.listdir(self.data_path):
            paths.append(self.data_path)
            snomed_releases.append(
                self._determine_release(self.data_path, strict=True))
            exts.append(self._determine_extension(self.data_path))
        else:
            for folder in os.listdir(self.data_path):
                if "SnomedCT" in folder:
                    paths.append(os.path.join(self.data_path, folder))
                    rel = self._determine_release(folder, strict=True)
                    snomed_releases.append(rel)
                    exts.append(self._determine_extension(paths[-1]))
        if len(paths) == 0:
            raise FileNotFoundError('Incorrect path to SNOMED CT directory')
        return paths, snomed_releases, exts

    def _refset_df2dict(self, refset_df: pd.DataFrame) -> dict:
        """
        This function takes a SNOMED refset DataFrame as an input and converts
        it into a dictionary.

        The DataFrame should contain the columns:
        'referencedComponentId','mapTarget','mapGroup','mapPriority','mapRule','mapAdvice'.

        Args:
            refset_df (pd.DataFrame) : DataFrame containing the refset data

        Returns:
            dict: mapping from SNOMED CT codes as key and the refset metadata
                list of dictionaries as values.
        """
        refset_dict = refset_df.groupby('referencedComponentId').apply(
            lambda group: [{'code': row['mapTarget'],
                            'mapGroup': row['mapPriority'],
                            'mapPriority': row['mapPriority'],
                            'mapRule': row['mapRule'],
                            'mapAdvice': row['mapAdvice']}
                           for _, row in group.iterrows()]).to_dict()
        return refset_dict

    def _map_snomed2refset(self):
        """
        Maps SNOMED CT concepts to refset mappings provided in the SNOMED CT
        release package.

        This function maps SNOMED CT concepts using the refset mappings in
        the Snapshot/Refset/Map directory.
        The refset mappings can either be ICD-10 codes in international
        releases or OPCS4 codes for SNOMED UK_extension, if available.

        Returns:
            pd.DataFrame: Dataframe containing SNOMED CT to refset mappings
                and metadata.
            OR
            tuple: Tuple of dataframes containing SNOMED CT to refset mappings
                and metadata (ICD-10, OPCS4), if uk_ext is True.
        """
        dfs2merge = []
        for i, snomed_release in enumerate(self.snomed_releases):
            self._set_extension(snomed_release, self.exts[i])
            refset_terminology = os.path.join(
                self.paths[i], PER_FILE_TYPE_PATHS[RefSetFileType.refset])
            icd10_ref_set = self._extension.value.exp_files.get_refset()
            if icd10_ref_set is None or _IGNORE_TAG in icd10_ref_set or (
                    self.bundle and self.bundle.value.has_invalid(
                        self._extension, [RefSetFileType.concept,
                                          RefSetFileType.description])):
                continue
            for f in os.listdir(refset_terminology):
                m = re.search(f'{icd10_ref_set}' + r'_(.*)_\d*.txt', f)
                if m:
                    snomed_v = m.group(1)
            mappings = parse_file(
                f'{refset_terminology}'
                f'/{icd10_ref_set}_{snomed_v}_{snomed_release}.txt')
            mappings = mappings[mappings.active == '1']
            icd_mappings = mappings.sort_values(
                by=['referencedComponentId', 'mapPriority',
                    'mapGroup']).reset_index(drop=True)
            dfs2merge.append(icd_mappings)
        mapping_df = pd.concat(dfs2merge)
        del dfs2merge
        if any(ext in (SupportedExtension.UK_CLINICAL,
                       SupportedExtension.UK_DRUG)
               for ext in self.exts):
            opcs_df = mapping_df[mapping_df['refsetId'] == self.opcs_refset_id]
            icd10_df = mapping_df[mapping_df['refsetId']
                                  == '999002271000000101']
            return icd10_df, opcs_df
        else:
            return mapping_df, None


class UnkownSnomedReleaseException(ValueError):

    def __init__(self, *args) -> None:
        super().__init__(*args)



import numpy as np
import logging
from typing import Type

from medcat.cdb import CDB
from medcat.vocab import Vocab


logger = logging.getLogger(__name__)


def calc_matrix(vocab: Vocab, target_size: int) -> np.ndarray:
    """Calculate the transformation matrix based on the word vectors in the
    Vocab.

    Performs Principal Component Analysis (PCA).
    This first means all the word vectors in the Vocab.
    It then finds the covariance matrix.
    After that, the eigenvalues and and eigenvectors are calculated.
    And the `target_size` eigenvectors corresponding to the largest
    eigenvalues are selected to create the transformation matrix.
    Args:
        vocab (Vocab): The Vocab.
        target_size (int): The target vector size.
    Returns:
        np.ndarray: The transformation matrix.
    """
    all_vecs = np.vstack(
        [value['vector'] for value in vocab.vocab.values()
         if value['vector'] is not None]
    )
    logger.debug("Vocab vectors have a total shape of %s", np.shape(all_vecs))
    all_vecs_meaned = all_vecs - np.mean(all_vecs, axis=0)
    cov_matrix = np.cov(all_vecs_meaned, rowvar=False)
    eigenvalues, eigenvectors = np.linalg.eigh(cov_matrix)
    sorted_idx = np.argsort(eigenvalues)[::-1]
    logger.debug("The sorted eigenvalues are as follows: %s",
                 [f"{v:5.2f}" for v in eigenvalues[sorted_idx]])
    sorted_eigenvectors = eigenvectors[:, sorted_idx]
    transformation_matrix = sorted_eigenvectors[:, :target_size]
    return transformation_matrix.T


def convert_vec(cur: np.ndarray, matrix: np.ndarray,
                target_dtype: Type = np.float32) -> np.ndarray:
    """Helper function to convert the vector.

    This also guarantees uniform typing (of np.float32) since in our
    experience some vectors may be of a different type before (i.e np.float64).

    Args:
        cur (np.ndarray): The current vector.
        matrix (np.ndarray): The transformation matrix.
        target_dtype (Type): The target element data ype.
            Defaults to np.float32.
    Returns:
        np.ndarray: The transformed vector.
    """
    return (matrix @ cur).astype(target_dtype)


def convert_vocab(vocab: Vocab, matrix: np.ndarray) -> None:
    """Use the transformation matrix to convert the word vectors.

    Args:
        vocab (Vocab): The Vocab.
        matrix (np.ndarray): The transformation matrix.
    """
    for d in vocab.vocab.values():
        cvec = d['vector']
        if cvec is None:
            continue
        d['vector'] = convert_vec(cvec, matrix)
    logger.info("Recalc cumulative sums (instead of unigram table)")
    vocab.init_cumsums()


def convert_context_vectors(cdb: CDB, matrix: np.ndarray) -> None:
    """Use the transformation matrix to convert the context vectors within the
    CDB.

    Args:
        cdb (CDB): The Context Database.
        matrix (np.ndarray): The transformation matrix.
    """
    for cuiinfo in cdb.cui2info.values():
        if 'context_vectors' not in cuiinfo:
            continue
        per_cui_dict = cuiinfo['context_vectors']
        if per_cui_dict is None:
            continue
        for type_name, cur_vec in list(per_cui_dict.items()):
            per_cui_dict[type_name] = convert_vec(cur_vec, matrix)
    cdb.is_dirty = True


def convert_vocab_vector_size(cdb: CDB, vocab: Vocab, vec_size: int):
    """Convert the vocab vector size to a smaller one.

    This uses Principal Component Analysis (PCA). The idea is that we
    first center all the word vectors (in Vocab), then compute the
    covariance matrix, then find the eigenvalues and eigenvectors,
    and then we select the top `vec_size` eigenvectors.
    This produces a transformation matrix of shape (vec_size, N),
    where N is the current vector length in the vocab.

    After that, we perform the transformation. First we transform all
    the vectors in the Vocab. And then we transform all the context
    vectors defined within the CDB.

    NOTE: This requires the CDB as well since the per concept context
    vectors stored within it are based on the vectors in the vocab and
    thus they also need to be transformed.

    Args:
        cdb (CDB): The Concept Database.
        vocab (Vocab): The Vocab.
        vec_size (int): The target vector size.
    """
    logger.info("Converting Vocab and CDB to size %s. Calculating "
                "transformation matrix", vec_size)
    matrix = calc_matrix(vocab, vec_size)
    logger.info("Found transformation matrix with shape %s. "
                "Now converting vocab.", matrix.shape)
    convert_vocab(vocab, matrix)
    logger.info("Done converting vocab, now converting the per concept "
                "context vectors defined in the CDB.")
    convert_context_vectors(cdb, matrix)
    logger.info("Done with the conversion to vocab vector size %s.",
                vec_size)



"""I would just ignore this whole class, it's just a lot of rules that work
nicely for CDB once the software is trained the main thing are the context
vectors.
"""
import logging
from typing import Optional
from medcat.tokenizing.tokens import (MutableEntity, MutableToken,
                                      MutableDocument)
from medcat.cdb import CDB
from medcat.config import Config
from medcat.tokenizing.tokenizers import BaseTokenizer

logger = logging.getLogger(__name__)


def annotate_name(tokenizer: BaseTokenizer, name: str,
                  tkns: list[MutableToken],
                  doc: MutableDocument, cdb: CDB,
                  label: str):
    entity: MutableEntity = tokenizer.create_entity(
        doc, tkns[0].base.index, tkns[-1].base.index + 1, label=label)
    # Only set this property when using a vocab approach
    # and where this name fits a name in the cdb.
    # All standard name entity recognition models will not set this.
    entity.detected_name = name
    entity.link_candidates = list(cdb.name2info[name]['per_cui_status'])
    entity.id = len(doc.ner_ents)
    entity.confidence = -1  # This does not calculate confidence
    # Append the entity to the document
    doc.ner_ents.append(entity)

    # Not necessary, but why not
    logger.debug("NER detected an entity.\n\tDetected name: %s" +
                 "\n\tLink candidates: %s\n", entity.detected_name,
                 entity.link_candidates)
    return entity


def maybe_annotate_name(tokenizer: BaseTokenizer, name: str,
                        tkns: list[MutableToken],
                        doc: MutableDocument, cdb: CDB, config: Config,
                        label: str = 'concept'
                        ) -> Optional[MutableEntity]:
    """Given a name it will check should it be annotated based on config rules.
    If yes the annotation will be added to the doc.entities array.

    Args:
        tokenizer (BaseTokenizer):
            The tokenizer (probably SpaCy).
        name (str):
            The name found in the text of the document.
        tkns (list[MutableToken]):
            Tokens that belong to this name in the spacy document.
        doc (BaseDocument):
            Spacy document to be annotated with named entities.
        cdb (CDB):
            Concept database.
        config (Config):
            Global config for medcat.
        label (str):
            Label for this name (usually `concept` if we are using
            a vocab based approach).

    Returns:
        Optional[BaseEntity]: The entity, if relevant.
    """

    logger.debug("Maybe annotating name: %s", name)

    # Check uppercase to distinguish uppercase and lowercase
    # words that have a different meaning.
    if config.components.ner.check_upper_case_names:
        # Check whether name is completely uppercase in CDB.
        is_upper = (cdb.name2info[name]['is_upper']
                    if name in cdb.name2info else False)
        if is_upper:
            # Check whether tokens are also in uppercase. If tokens
            # are not in uppercase, there is a mismatch.
            if not all([x.base.is_upper for x in tkns]):
                return None

    if len(name) >= config.components.ner.min_name_len:
        # Check the upper case limit, last part checks if it is
        # one token and uppercase
        if (len(name) >= config.components.ner.upper_case_limit_len or
                (len(tkns) == 1 and tkns[0].base.is_upper)):
            # Everything is fine, mark name
            return annotate_name(tokenizer, name, tkns, doc, cdb, label)

    return None



import torch
from collections import OrderedDict
from typing import Optional, Any, Iterable
from torch import nn, Tensor
from transformers import BertModel, AutoConfig
from medcat.config.config_meta_cat import ConfigMetaCAT
import logging
logger = logging.getLogger(__name__)


class LSTM(nn.Module):
    def __init__(self, embeddings: Optional[Tensor], config: ConfigMetaCAT
                 ) -> None:
        super(LSTM, self).__init__()

        self.config = config
        # Get the required sizes
        vocab_size = config.general.vocab_size
        embedding_size = config.model.input_size

        # Initialize embeddings
        self.embeddings = nn.Embedding(vocab_size, embedding_size,
                                       padding_idx=config.model.padding_idx)
        if embeddings is not None:
            self.embeddings.load_state_dict(OrderedDict(
                [('weight', embeddings)]))
            # Disable training for the embeddings - IMPORTANT
            self.embeddings.weight.requires_grad = config.model.emb_grad

        # Create the RNN cell - divide
        self.rnn = nn.LSTM(
            input_size=config.model.input_size,
            hidden_size=(config.model.hidden_size //
                         config.model.num_directions),
            num_layers=config.model.num_layers,
            dropout=config.model.dropout,
            bidirectional=config.model.num_directions == 2)
        self.fc1 = nn.Linear(config.model.hidden_size, config.model.nclasses)

        self.d1 = nn.Dropout(config.model.dropout)

    def forward(self,
                input_ids: torch.LongTensor,
                center_positions: Tensor,
                attention_mask: Optional[torch.FloatTensor] = None,
                ignore_cpos: bool = False) -> Tensor:
        x = input_ids
        # Get the mask from x
        if attention_mask is None:
            mask = x != self.config.model.padding_idx
        else:
            mask = attention_mask

        # Embed the input: from id -> vec
        # x.shape = batch_size x sequence_length x emb_size
        x = self.embeddings(x)

        # Tell RNN to ignore padding and set the batch_first to True
        x2 = nn.utils.rnn.pack_padded_sequence(
            x, mask.sum(1).int().view(-1).cpu(), batch_first=True,
            enforce_sorted=False)

        # Run 'x' through the RNN
        x3, hidden = self.rnn(x2)

        # Add the padding again
        x4, _ = torch.nn.utils.rnn.pad_packed_sequence(x3, batch_first=True)

        # Get what we need
        # row_indices = torch.arange(0, x.size(0)).long()

        # If this is  True we will always take the last state and not CPOS
        if ignore_cpos:
            x4 = hidden[0]
            x4 = x4.view(
                self.config.model.num_layers,
                self.config.model.num_directions, -1,
                (self.config.model.hidden_size //
                 self.config.model.num_directions))
            x4 = x4[-1, :, :, :].permute(1, 2, 0).reshape(
                -1, self.config.model.hidden_size)
        else:
            x_all = []
            for i, indices in enumerate(center_positions):
                this_hidden = x4[i, indices, :]
                to_append, _ = torch.max(this_hidden, dim=0)
                x_all.append(to_append)

            x4 = torch.stack(x_all)

        # Push x through the fc network and add dropout
        x5 = self.d1(x4)
        x6 = self.fc1(x5)

        return x6


class BertForMetaAnnotation(nn.Module):
    _keys_to_ignore_on_load_unexpected: list[str] = [r"pooler"]  # type: ignore

    def __init__(self, config: ConfigMetaCAT):
        super(BertForMetaAnnotation, self).__init__()
        _bertconfig = AutoConfig.from_pretrained(
            config.model.model_variant,
            num_hidden_layers=config.model.num_layers)
        if config.model.input_size != _bertconfig.hidden_size:
            logger.warning(
                "Input size for %s model should be %d, provided input size is "
                "%d. Input size changed to %d", config.model.model_variant,
                _bertconfig.hidden_size, config.model.input_size,
                _bertconfig.hidden_size)

        bert = BertModel.from_pretrained(config.model.model_variant,
                                         config=_bertconfig)
        self.config = config
        self.bert = bert
        self.num_labels = config.model.nclasses
        for param in self.bert.parameters():
            param.requires_grad = not config.model.model_freeze_layers

        hidden_size_2 = int(config.model.hidden_size / 2)
        # dropout layer
        self.dropout = nn.Dropout(config.model.dropout)
        # relu activation function
        self.relu = nn.ReLU()
        # dense layer 1
        self.fc1 = nn.Linear(_bertconfig.hidden_size * 2,
                             config.model.hidden_size)
        # dense layer 2
        self.fc2 = nn.Linear(config.model.hidden_size, hidden_size_2)
        # dense layer 3
        self.fc3 = nn.Linear(hidden_size_2, hidden_size_2)
        # dense layer 3 (Output layer)
        model_arch_config = config.model.model_architecture_config

        if (model_arch_config['fc3'] is True and
                model_arch_config['fc2'] is False):
            logger.warning("FC3 can only be used if FC2 is also enabled. "
                           "Enabling FC2...")
            config.model.model_architecture_config['fc2'] = True

        if model_arch_config is not None:
            if model_arch_config['fc2'] is True:
                self.fc4 = nn.Linear(hidden_size_2, self.num_labels)
            else:
                self.fc4 = nn.Linear(config.model.hidden_size, self.num_labels)
        else:
            self.fc4 = nn.Linear(hidden_size_2, self.num_labels)
        # softmax activation function
        self.softmax = nn.LogSoftmax(dim=1)

    def forward(
            self,
            input_ids: Optional[torch.LongTensor] = None,
            attention_mask: Optional[torch.FloatTensor] = None,
            token_type_ids: Optional[torch.LongTensor] = None,
            position_ids: Optional[torch.LongTensor] = None,
            head_mask: Optional[torch.FloatTensor] = None,
            inputs_embeds: Optional[torch.FloatTensor] = None,
            labels: Optional[torch.LongTensor] = None,
            center_positions: Iterable[Any] = [],
            ignore_cpos: Optional[bool] = None,
            output_attentions: Optional[bool] = None,
            output_hidden_states: Optional[bool] = None,
            return_dict: Optional[bool] = None
    ):
        """

        Args:
            input_ids (Optional[torch.LongTensor]):
                The input IDs. Defaults to None.
            attention_mask (Optional[torch.FloatTensor]):
                The attention mask. Defaults to None.
            token_type_ids (Optional[torch.LongTensor]):
                Type IDs of the tokens. Defaults to None.
            position_ids (Optional[torch.LongTensor]):
                Position IDs. Defaults to None.
            head_mask (Optional[torch.FloatTensor]):
                Head mask. Defaults to None.
            inputs_embeds (Optional[torch.FloatTensor]):
                Input embeddings. Defaults to None.
            labels (Optional[torch.LongTensor]):
                Labels for computing the token classification loss. Indices
                should be in `[0, ..., config.num_labels - 1]`.
                Defaults to None.
            center_positions (Optional[Any]): Cennter positions.
                Defaults to None.
            output_attentions (Optional[bool]): Output attentions.
                Defaults to None.
            ignore_cpos:
                If center positions are to be ignored.
            output_hidden_states (Optional[bool]):
                Output hidden states. Defaults to None.
            return_dict (Optional[bool]):
                Whether to return a dict. Defaults to None.

        Returns:
            TokenClassifierOutput: The token classifier output.
        """

        outputs = self.bert(  # type: ignore
            input_ids,
            attention_mask=attention_mask, output_hidden_states=True
        )

        x_all = []
        for i, indices in enumerate(center_positions):
            this_hidden: torch.Tensor = outputs.last_hidden_state[
                i, indices, :]
            to_append, _ = torch.max(this_hidden, dim=0)
            x_all.append(to_append)

        x = torch.stack(x_all)

        pooled_output = outputs[1]
        x = torch.cat((x, pooled_output), dim=1)

        # fc1
        x = self.dropout(x)
        x = self.fc1(x)
        x = self.relu(x)

        if self.config.model.model_architecture_config is not None:
            if self.config.model.model_architecture_config['fc2'] is True:
                # fc2
                x = self.fc2(x)
                x = self.relu(x)
                x = self.dropout(x)

                if self.config.model.model_architecture_config['fc3'] is True:
                    # fc3
                    x = self.fc3(x)
                    x = self.relu(x)
                    x = self.dropout(x)
        else:
            # fc2
            x = self.fc2(x)
            x = self.relu(x)
            x = self.dropout(x)

            # fc3
            x = self.fc3(x)
            x = self.relu(x)
            x = self.dropout(x)

        # output layer
        x = self.fc4(x)
        return x



import dill


class RemappingUnpickler(dill.Unpickler):
    # TODO: remove at some point - this only serves to load beta models
    def find_class(self, module: str, name: str):
        # Remap old module path to new one
        module, name = fix_module_and_cls_name(module, name)
        return super().find_class(module, name)


def fix_module_and_cls_name(module_name: str, cls_name: str
                            ) -> tuple[str, str]:
    if module_name.startswith("medcat2"):
        module_name = module_name.replace("medcat2", "medcat", 1)
    return module_name, cls_name



from typing import Optional, Union, Any, overload, Literal, Iterable, Iterator
from typing import cast
import os
import json
from datetime import date
from concurrent.futures import ProcessPoolExecutor, as_completed, Future
import itertools
from contextlib import contextmanager

import shutil
import logging

from medcat.utils.defaults import DEFAULT_PACK_NAME, COMPONENTS_FOLDER
from medcat.cdb import CDB
from medcat.vocab import Vocab
from medcat.config.config import Config, get_important_config_parameters
from medcat.trainer import Trainer
from medcat.storage.serialisers import serialise, AvailableSerialisers
from medcat.storage.serialisers import deserialise
from medcat.storage.serialisables import AbstractSerialisable
from medcat.utils.fileutils import ensure_folder_if_parent
from medcat.utils.hasher import Hasher
from medcat.pipeline.pipeline import Pipeline
from medcat.tokenizing.tokens import MutableDocument, MutableEntity
from medcat.tokenizing.tokenizers import SaveableTokenizer, TOKENIZER_PREFIX
from medcat.data.entities import Entity, Entities, OnlyCUIEntities
from medcat.data.model_card import ModelCard
from medcat.components.types import AbstractCoreComponent, HashableComponet
from medcat.components.addons.addons import AddonComponent
from medcat.utils.legacy.identifier import is_legacy_model_pack
from medcat.utils.defaults import avoid_legacy_conversion
from medcat.utils.defaults import doing_legacy_conversion_message
from medcat.utils.defaults import LegacyConversionDisabledError
from medcat.utils.usage_monitoring import UsageMonitor, _NoDelUM
from medcat.utils.import_utils import MissingDependenciesError


logger = logging.getLogger(__name__)


class CAT(AbstractSerialisable):
    """This is a collection of serialisable model parts.
    """
    FORCE_SPAWN_MP = True

    def __init__(self,
                 cdb: CDB,
                 vocab: Union[Vocab, None] = None,
                 config: Optional[Config] = None,
                 model_load_path: Optional[str] = None,
                 ) -> None:
        self.cdb = cdb
        self.vocab = vocab
        # ensure  config
        if config is None and self.cdb.config is None:
            raise ValueError("Need to specify a config for either CDB or CAT")
        elif config is None:
            config = cdb.config
        elif config is not None:
            self.cdb.config = config
        self.config = config

        self._trainer: Optional[Trainer] = None
        self._pipeline = self._recreate_pipe(model_load_path)
        self.usage_monitor = UsageMonitor(
            self._get_hash, self.config.general.usage_monitor)

    def _recreate_pipe(self, model_load_path: Optional[str] = None
                       ) -> Pipeline:
        if hasattr(self, "_pipeline"):
            old_pipe = self._pipeline
        else:
            old_pipe = None
        self._pipeline = Pipeline(self.cdb, self.vocab, model_load_path,
                                  old_pipe=old_pipe)
        return self._pipeline

    @classmethod
    def get_init_attrs(cls) -> list[str]:
        return ['cdb', 'vocab']

    @classmethod
    def ignore_attrs(cls) -> list[str]:
        return [
            '_trainer',  # recreate if nededed
            '_pipeline',  # need to recreate regardless
            'config',  # will be loaded along with CDB
            'usage_monitor',  # will be created at startup
        ]

    def __call__(self, text: str) -> Optional[MutableDocument]:
        doc = self._pipeline.get_doc(text)
        if self.usage_monitor.should_monitor:
            self.usage_monitor.log_inference(len(text), len(doc.linked_ents))
        return doc

    def _ensure_not_training(self) -> None:
        """Method to ensure config is not set to train.

        `config.components.linking.train` should only be True while training
        and not during inference.
        This aalso corrects the setting if necessary.
        """
        # pass
        if self.config.components.linking.train:
            logger.warning("Training was enabled during inference. "
                           "It was automatically disabled.")
            self.config.components.linking.train = False

    @overload
    def get_entities(self,
                     text: str,
                     only_cui: Literal[False] = False,
                     # TODO : addl_info
                     ) -> Entities:
        pass

    @overload
    def get_entities(self,
                     text: str,
                     only_cui: Literal[True] = True,
                     # TODO : addl_info
                     ) -> OnlyCUIEntities:
        pass

    @overload
    def get_entities(self,
                     text: str,
                     only_cui: bool = False,
                     # TODO : addl_info
                     ) -> Union[dict, Entities, OnlyCUIEntities]:
        pass

    def get_entities(self,
                     text: str,
                     only_cui: bool = False,
                     # TODO : addl_info
                     ) -> Union[dict, Entities, OnlyCUIEntities]:
        """Get the entities recognised and linked within the provided text.

        This will run the text through the pipeline and annotated the
        recognised and linked entities.

        Args:
            text (str): The text to use.
            only_cui (bool, optional): Whether to only output the CUIs
                rather than the entire context. Defaults to False.

        Returns:
            Union[dict, Entities, OnlyCUIEntities]: The entities found and
                linked within the text.
        """
        self._ensure_not_training()
        doc = self(text)
        if not doc:
            return {}
        return self._doc_to_out(doc, only_cui=only_cui)

    def _mp_worker_func(
            self,
            texts_and_indices: list[tuple[str, str, bool]]
            ) -> list[tuple[str, str, Union[dict, Entities, OnlyCUIEntities]]]:
        # NOTE: this is needed for subprocess as otherwise they wouldn't have
        #       any of these set
        # NOTE: these need to by dynamic in case the extra's aren't included
        try:
            from medcat.components.addons.meta_cat import MetaCATAddon
            has_meta_cat = True
        except MissingDependenciesError:
            has_meta_cat = False
        try:
            from medcat.components.addons.relation_extraction.rel_cat import (
                RelCATAddon)
            has_rel_cat = True
        except MissingDependenciesError:
            has_rel_cat = False
        for addon in self._pipeline.iter_addons():
            if has_meta_cat and isinstance(addon, MetaCATAddon):
                addon._init_data_paths(self._pipeline.tokenizer)
            elif has_rel_cat and isinstance(addon, RelCATAddon):
                addon._rel_cat._init_data_paths()
        return [
            (text, text_index, self.get_entities(text, only_cui=only_cui))
            for text, text_index, only_cui in texts_and_indices]

    def _generate_batches_by_char_length(
            self,
            text_iter: Union[Iterator[str], Iterator[tuple[str, str]]],
            batch_size_chars: int,
            only_cui: bool,
            ) -> Iterator[list[tuple[str, str, bool]]]:
        docs: list[tuple[str, str, bool]] = []
        char_count = 0
        for i, _doc in enumerate(text_iter):
            # NOTE: not sure why mypy is complaining here
            doc = cast(
                str, _doc[1] if isinstance(_doc, tuple) else _doc)
            doc_index: str = _doc[0] if isinstance(_doc, tuple) else str(i)
            clen = len(doc)
            char_count += clen
            if char_count > batch_size_chars:
                yield docs
                docs = []
                char_count = clen
            docs.append((doc, doc_index, only_cui))

        if len(docs) > 0:
            yield docs

    def _generate_batches(
            self,
            text_iter: Union[Iterator[str], Iterator[tuple[str, str]]],
            batch_size: int,
            batch_size_chars: int,
            only_cui: bool,
            ) -> Iterator[list[tuple[str, str, bool]]]:
        if batch_size_chars < 1 and batch_size < 1:
            raise ValueError("Either `batch_size` or `batch_size_chars` "
                             "must be greater than 0.")
        if batch_size > 0 and batch_size_chars > 0:
            raise ValueError(
                "Cannot specify both `batch_size` and `batch_size_chars`. "
                "Please use one of them.")
        if batch_size_chars > 0:
            return self._generate_batches_by_char_length(
                text_iter, batch_size_chars, only_cui)
        else:
            return self._generate_simple_batches(
                text_iter, batch_size, only_cui)

    def _generate_simple_batches(
            self,
            text_iter: Union[Iterator[str], Iterator[tuple[str, str]]],
            batch_size: int,
            only_cui: bool,
            ) -> Iterator[list[tuple[str, str, bool]]]:
        text_index = 0
        while True:
            # Take a small batch from the iterator
            batch = list(itertools.islice(text_iter, batch_size))
            if not batch:
                break
            # NOTE: typing is correct:
            #        - if str, then (str, int, bool)
            #        - if tuple, then (str, int, bool)
            #       but for some reason mypy complains
            yield [
                (text, str(text_index + i), only_cui)  # type: ignore
                if isinstance(text, str) else
                (text[1], text[0], only_cui)
                for i, text in enumerate(batch)
            ]
            text_index += len(batch)

    def _mp_one_batch_per_process(
            self,
            executor: ProcessPoolExecutor,
            batch_iter: Iterator[list[tuple[str, str, bool]]],
            external_processes: int
            ) -> Iterator[tuple[str, Union[dict, Entities, OnlyCUIEntities]]]:
        futures: list[Future] = []
        # submit batches, one for each external processes
        for _ in range(external_processes):
            try:
                batch = next(batch_iter)
                futures.append(
                    executor.submit(self._mp_worker_func, batch))
            except StopIteration:
                break
        if not futures:
            # NOTE: if there wasn't any data, we didn't process anything
            return
        # Main process works on next batch while workers are busy
        main_batch: Optional[list[tuple[str, str, bool]]]
        try:
            main_batch = next(batch_iter)
            main_results = self._mp_worker_func(main_batch)

            # Yield main process results immediately
            for result in main_results:
                yield result[1], result[2]

        except StopIteration:
            main_batch = None
        # since the main process did around the same amount of work
        # we would expect all subprocess to have finished by now
        # so we're going to wait for them to finish, yield their results,
        # and subsequently submit the next batch to keep them busy
        for _ in range(external_processes):
            if not futures:
                # NOTE: if there's no futures then there can't be
                #       anything to batch
                break
            # Wait for any future to complete
            done_future = next(as_completed(futures))
            futures.remove(done_future)

            # Yield all results from this batch
            for result in done_future.result():
                yield result[1], result[2]

            # Submit next batch to keep workers busy
            try:
                batch = next(batch_iter)
                futures.append(
                    executor.submit(self._mp_worker_func, batch))
            except StopIteration:
                # NOTE: if there's nothing to batch, we've got nothing
                #       to submit in terms of new work to the workers,
                #       but we may still have some futures to wait for
                pass

    def get_entities_multi_texts(
            self,
            texts: Union[Iterable[str], Iterable[tuple[str, str]]],
            only_cui: bool = False,
            n_process: int = 1,
            batch_size: int = -1,
            batch_size_chars: int = 1_000_000,
            ) -> Iterator[tuple[str, Union[dict, Entities, OnlyCUIEntities]]]:
        """Get entities from multiple texts (potentially in parallel).

        If `n_process` > 1, `n_process - 1` new processes will be created
        and data will be processed on those as well as the main process in
        parallel.

        Args:
            texts (Union[Iterable[str], Iterable[tuple[str, str]]]):
                The input text. Either an iterable of raw text or one
                with in the format of `(text_index, text)`.
            only_cui (bool):
                Whether to only return CUIs rather than other information
                like start/end and annotated value. Defaults to False.
            n_process (int):
                Number of processes to use. Defaults to 1.
            batch_size (int):
                The number of texts to batch at a time. A batch of the
                specified size will be given to each worker process.
                Defaults to -1 and in this case the character count will
                be used instead.
            batch_size_chars (int):
                The maximum number of characters to process in a batch.
                Each process will be given batch of texts with a total
                number of characters not exceeding this value. Defaults
                to 1,000,000 characters. Set to -1 to disable.

        Yields:
            Iterator[tuple[str, Union[dict, Entities, OnlyCUIEntities]]]:
                The results in the format of (text_index, entities).
        """
        text_iter = cast(
            Union[Iterator[str], Iterator[tuple[str, str]]], iter(texts))
        batch_iter = self._generate_batches(
            text_iter, batch_size, batch_size_chars, only_cui)
        if n_process == 1:
            # just do in series
            for batch in batch_iter:
                for _, text_index, result in self._mp_worker_func(batch):
                    yield text_index, result
            return

        with self._no_usage_monitor_exit_flushing():
            yield from self._multiprocess(n_process, batch_iter)

    @contextmanager
    def _no_usage_monitor_exit_flushing(self):
        # NOTE: the `UsageMonitor.__del__` method can cause
        #       multiprocessing to stall while it waits for it to be
        #       called. So here we remove the method.
        #       However, due to the object being pickled for multiprocessing
        #       purposes, the class'es `__del__` method will be used anyway.
        #       So we need to trick it into using a different class.
        original_cls = self.usage_monitor.__class__
        self.usage_monitor.__class__ = _NoDelUM
        try:
            yield
        finally:
            self.usage_monitor.__class__ = original_cls

    def _multiprocess(
            self, n_process: int,
            batch_iter: Iterator[list[tuple[str, str, bool]]]
            ) -> Iterator[tuple[str, Union[dict, Entities, OnlyCUIEntities]]]:
        external_processes = n_process - 1
        if self.FORCE_SPAWN_MP:
            import multiprocessing as mp
            logger.info(
                "Forcing multiprocessing start method to 'spawn' "
                "due to known compatibility issues with 'fork' and "
                "libraries using threads or native extensions.")
            mp.set_start_method("spawn", force=True)
        with ProcessPoolExecutor(max_workers=external_processes) as executor:
            yield from self._mp_one_batch_per_process(
                executor, batch_iter, external_processes)

    def _get_entity(self, ent: MutableEntity,
                    doc_tokens: list[str],
                    cui: str) -> Entity:
        context_left = self.config.annotation_output.context_left
        context_right = self.config.annotation_output.context_right

        if context_left > 0 and context_right > 0:
            left_s = max(ent.base.start_index - context_left, 0)
            left_e = ent.base.start_index
            left_context = doc_tokens[left_s:left_e]
            right_s = ent.base.end_index
            right_e = min(ent.base.end_index + context_right, len(doc_tokens))
            right_context = doc_tokens[right_s:right_e]
            ent_s, ent_e = ent.base.start_index, ent.base.end_index
            center_context = doc_tokens[ent_s:ent_e]
        else:
            left_context = []
            right_context = []
            center_context = []

        # NOTE: in case the CUI is not in the CDB, we don't want to fail here
        def_ci: dict[str, list[str]] = {'type_ids': []}
        out_dict: Entity = {
            'pretty_name': self.cdb.get_name(cui),
            'cui': cui,
            'type_ids': list(self.cdb.cui2info.get(cui, def_ci)['type_ids']),
            'source_value': ent.base.text,
            'detected_name': str(ent.detected_name),
            'acc': ent.context_similarity,
            'context_similarity': ent.context_similarity,
            'start': ent.base.start_char_index,
            'end': ent.base.end_char_index,
            # TODO: add additional info (i.e mappings)
            # for addl in addl_info:
            #     tmp = self.cdb.addl_info.get(addl, {}).get(cui, [])
            #     out_ent[addl.split("2")[-1]] = list(tmp) if type(tmp) is
            # set else tmp
            'id': ent.id,
            # TODO: add met annotations
            # if hasattr(ent._, 'meta_anns') and ent._.meta_anns:
            #     out_ent['meta_anns'] = ent._.meta_anns
            'meta_anns': {},
            'context_left': left_context,
            'context_center': center_context,
            'context_right': right_context,
        }
        # addons:
        out_dict.update(self.get_addon_output(ent))  # type: ignore
        return out_dict

    def get_addon_output(self, ent: MutableEntity) -> dict[str, dict]:
        """Get the addon output for the entity.

        This includes a key-value pair for each addon that provides some.
        Sometimes same-type addons may combine their output under the same key.

        Args:
            ent (MutableEntity): The entity in quesiton.

        Raises:
            ValueError: If unable to merge multiple addon output.

        Returns:
            dict[str, dict]: All the addon output.
        """
        out_dict: dict[str, dict] = {}
        for addon in self._pipeline._addons:
            if not addon.include_in_output:
                continue
            key, val = addon.get_output_key_val(ent)
            if key in out_dict:
                # e.g multiple meta_anns types
                # NOTE: type-ignore due to the strict TypedDict implementation
                cur_val = out_dict[key]  # type: ignore
                if not isinstance(cur_val, dict):
                    raise ValueError(
                        "Unable to merge multiple addon output for the same "
                        f" key. Tried to update '{key}'. Previously had "
                        f"{cur_val}, got {val} from addon {addon.full_name}")
                cur_val.update(val)
            else:
                # NOTE: type-ignore due to the strict TypedDict implementation
                out_dict[key] = val  # type: ignore
        return out_dict

    def _doc_to_out_entity(self, ent: MutableEntity,
                           doc_tokens: list[str],
                           only_cui: bool,
                           ) -> tuple[int, Union[Entity, str]]:
        cui = str(ent.cui)
        if not only_cui:
            out_ent = self._get_entity(ent, doc_tokens, cui)
            return ent.id, out_ent
        else:
            return ent.id, cui

    def _doc_to_out(self,
                    doc: MutableDocument,
                    only_cui: bool,
                    # addl_info: list[str], # TODO
                    out_with_text: bool = False
                    ) -> Union[Entities, OnlyCUIEntities]:
        out: Union[Entities, OnlyCUIEntities] = {'entities': {},
                                                 'tokens': []}  # type: ignore
        cnf_annotation_output = self.config.annotation_output
        _ents = doc.linked_ents

        if cnf_annotation_output.lowercase_context:
            doc_tokens = [tkn.base.text_with_ws.lower() for tkn in list(doc)]
        else:
            doc_tokens = [tkn.base.text_with_ws for tkn in list(doc)]

        for _, ent in enumerate(_ents):
            ent_id, ent_dict = self._doc_to_out_entity(ent, doc_tokens,
                                                       only_cui)
            # NOTE: the types match - not sure why mypy is having issues
            out['entities'][ent_id] = ent_dict  # type: ignore

        if cnf_annotation_output.include_text_in_output or out_with_text:
            out['text'] = doc.base.text
        return out

    @property
    def trainer(self):
        """The trainer object."""
        if not self._trainer:
            self._trainer = Trainer(self.cdb, self.__call__, self._pipeline)
        return self._trainer

    def save_model_pack(
            self, target_folder: str, pack_name: str = DEFAULT_PACK_NAME,
            serialiser_type: Union[str, AvailableSerialisers] = 'dill',
            make_archive: bool = True,
            only_archive: bool = False,
            add_hash_to_pack_name: bool = True,
            change_description: Optional[str] = None,
            ) -> str:
        """Save model pack.

        The resulting model pack name will have the hash of the model pack
        in its name if (and only if) the default model pack name is used.

        Args:
            target_folder (str):
                The folder to save the pack in.
            pack_name (str, optional): The model pack name.
                Defaults to DEFAULT_PACK_NAME.
            serialiser_type (Union[str, AvailableSerialisers], optional):
                The serialiser type. Defaults to 'dill'.
            make_archive (bool):
                Whether to make the arhive /.zip file. Defaults to True.
            only_archive (bool):
                Whether to clear the non-compressed folder. Defaults to False.
            add_hash_to_pack_name (bool):
                Whether to add the hash to the pack name. This is only relevant
                if pack_name is specified. Defaults to True.
            change_description (Optional[str]):
                If provided, this the description will be added to the
                model description. Defaults to None.

        Returns:
            str: The final model pack path.
        """
        self.config.meta.mark_saved_now()
        # figure out the location/folder of the saved files
        hex_hash = self._versioning(change_description)
        if pack_name == DEFAULT_PACK_NAME or add_hash_to_pack_name:
            pack_name = f"{pack_name}_{hex_hash}"
        model_pack_path = os.path.join(target_folder, pack_name)
        # ensure target folder and model pack folder exist
        ensure_folder_if_parent(model_pack_path)
        # tokenizer (e.g spacy model) - needs to saved before since
        #     it changes config slightly
        if isinstance(self._pipeline.tokenizer, SaveableTokenizer):
            internals_path = self._pipeline.tokenizer.save_internals_to(
                model_pack_path)
            self.config.general.nlp.modelname = internals_path
        # serialise
        serialise(serialiser_type, self, model_pack_path)
        model_card: str = self.get_model_card(as_dict=False)
        model_card_path = os.path.join(model_pack_path, "model_card.json")
        with open(model_card_path, 'w') as f:
            f.write(model_card)
        # components
        components_folder = os.path.join(
            model_pack_path, COMPONENTS_FOLDER)
        self._pipeline.save_components(serialiser_type, components_folder)
        # zip everything
        if make_archive:
            shutil.make_archive(model_pack_path, 'zip',
                                root_dir=model_pack_path)
            if only_archive:
                logger.info("Removing the non-archived model pack folder: %s",
                            model_pack_path)
                shutil.rmtree(model_pack_path, ignore_errors=True)
                # change the model pack path to the zip file so that we
                # refer to an existing file
                model_pack_path += ".zip"
        return model_pack_path

    def _get_hash(self) -> str:
        hasher = Hasher()
        logger.debug("Hashing the CDB")
        hasher.update(self.cdb.get_hash())
        for component in self._pipeline.iter_all_components():
            if isinstance(component, HashableComponet):
                logger.debug("Hashing for component %s",
                             type(component).__name__)
                hasher.update(component.get_hash())
        hex_hash = self.config.meta.hash = hasher.hexdigest()
        return hex_hash

    def _versioning(self, change_description: Optional[str]) -> str:
        date_today = date.today().strftime("%d %B %Y")
        if change_description is not None:
            self.config.meta.description += (
                f"\n[{date_today}] {change_description}")
        hex_hash = self._get_hash()
        history = self.config.meta.history
        if not history or history[-1] != hex_hash:
            history.append(hex_hash)
        logger.info("Got hash: %s", hex_hash)
        return hex_hash

    @classmethod
    def attempt_unpack(cls, zip_path: str) -> str:
        """Attempt unpack the zip to a folder and get the model pack path.

        If the folder already exists, no unpacking is done.

        Args:
            zip_path (str): The ZIP path

        Returns:
            str: The model pack path
        """
        base_dir = os.path.dirname(zip_path)
        filename = os.path.basename(zip_path)

        foldername = filename.replace(".zip", '')

        model_pack_path = os.path.join(base_dir, foldername)
        if os.path.exists(model_pack_path):
            logger.info(
                "Found an existing unzipped model pack at: %s, "
                "the provided zip will not be touched.", model_pack_path)
        else:
            logger.info("Unziping the model pack and loading models.")
            shutil.unpack_archive(zip_path, extract_dir=model_pack_path)
        return model_pack_path

    @classmethod
    def load_model_pack(cls, model_pack_path: str) -> 'CAT':
        """Load the model pack from file.

        Args:
            model_pack_path (str): The model pack path.

        Raises:
            ValueError: If the saved data does not represent a model pack.

        Returns:
            CAT: The loaded model pack.
        """
        if model_pack_path.endswith(".zip"):
            model_pack_path = cls.attempt_unpack(model_pack_path)
        logger.info("Attempting to load model from file: %s",
                    model_pack_path)
        is_legacy = is_legacy_model_pack(model_pack_path)
        avoid_legacy = avoid_legacy_conversion()
        if is_legacy and not avoid_legacy:
            from medcat.utils.legacy.conversion_all import Converter
            doing_legacy_conversion_message(logger, 'CAT', model_pack_path)
            return Converter(model_pack_path, None).convert()
        elif is_legacy and avoid_legacy:
            raise LegacyConversionDisabledError("CAT")
        # NOTE: ignoring addons since they will be loaded later / separately
        cat = deserialise(model_pack_path, model_load_path=model_pack_path,
                          ignore_folders_prefix={
                            AddonComponent.NAME_PREFIX,
                            # NOTE: will be loaded manually
                            AbstractCoreComponent.NAME_PREFIX,
                            # tokenizer stuff internals are loaded separately
                            # if appropraite
                            TOKENIZER_PREFIX,
                            # components will be loaded semi-manually
                            # within the creation of pipe
                            COMPONENTS_FOLDER})
        # NOTE: deserialising of components that need serialised
        #       will be dealt with upon pipeline creation automatically
        if not isinstance(cat, CAT):
            raise ValueError(f"Unable to load CAT. Got: {cat}")
        return cat

    @classmethod
    def load_cdb(cls, model_pack_path: str) -> CDB:
        """
        Loads the concept database from the provided model pack path

        Args:
            model_pack_path (str): path to model pack, zip or dir.

        Returns:
            CDB: The loaded concept database
        """
        zip_path = (model_pack_path if model_pack_path.endswith(".zip")
                    else model_pack_path + ".zip")
        model_pack_path = cls.attempt_unpack(zip_path)
        cdb_path = os.path.join(model_pack_path, "cdb")
        cdb = CDB.load(cdb_path)
        return cdb

    @classmethod
    def load_addons(
            cls, model_pack_path: str, meta_cat_config_dict: Optional[dict] = None
            ) -> list[tuple[str, AddonComponent]]:
        """Load addons based on a model pack path.

        Args:
            model_pack_path (str): path to model pack, zip or dir.
            meta_cat_config_dict (Optional[dict]):
                A config dict that will overwrite existing configs in meta_cat.
                e.g. meta_cat_config_dict = {'general': {'device': 'cpu'}}.
                Defaults to None.

        Returns:
            List[tuple(str, AddonComponent)]: list of pairs of adddon names the addons.
        """
        components_folder = os.path.join(model_pack_path, COMPONENTS_FOLDER)
        if not os.path.exists(components_folder):
            return []
        addon_paths = [
            folder_path
            for folder_name in os.listdir(components_folder)
            if os.path.isdir(folder_path := os.path.join(
                components_folder, folder_name))
            and folder_name.startswith(AddonComponent.NAME_PREFIX)
        ]
        loaded_addons = [
            addon for addon_path in addon_paths
            if isinstance(addon := deserialise(addon_path), AddonComponent)
        ]
        return [(addon.full_name, addon) for addon in loaded_addons]


    @overload
    def get_model_card(self, as_dict: Literal[True]) -> ModelCard:
        pass

    @overload
    def get_model_card(self, as_dict: Literal[False]) -> str:
        pass

    def get_model_card(self, as_dict: bool = False) -> Union[str, ModelCard]:
        """Get the model card either a (nested) `dict` or a json string.

        Args:
            as_dict (bool): Whether to return as dict. Defaults to False.

        Returns:
            Union[str, ModelCard]: The model card.
        """
        meta_cat_categories = [
            cnf.general.category_name  # type: ignore
            for cnf in self.config.components.addons
            if cnf.comp_name == 'meta_cat' and
            # NOTE: not the best way to check this,
            #       but I don't want to import the addon config
            type(cnf).__name__ == 'ConfigMetaCAT']
        cdb_info = self.cdb.get_basic_info()
        model_card: ModelCard = {
            'Model ID': self.config.meta.hash,
            'Last Modified On': self.config.meta.last_saved.isoformat(),
            'History (from least to most recent)': self.config.meta.history,
            'Description': self.config.meta.description,
            'Source Ontology': self.config.meta.ontology,
            'Location': self.config.meta.location,
            'MetaCAT models': meta_cat_categories,
            'Basic CDB Stats': cdb_info,
            'Performance': {},  # TODO
            'Important Parameters (Partial view, '
            'all available in cat.config)': get_important_config_parameters(
                self.config),
            'MedCAT Version': self.config.meta.medcat_version,
        }
        if as_dict:
            return model_card
        return json.dumps(model_card, indent=2, sort_keys=False)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, CAT):
            return False
        return (self.cdb == other.cdb and
                ((self.vocab is None and other.vocab is None)
                 or self.vocab == other.vocab))

    # addon (e.g MetaCAT) related stuff

    def add_addon(self, addon: AddonComponent) -> None:
        self.config.components.addons.append(addon.config)
        self._pipeline.add_addon(addon)



from typing import Protocol, Optional, Iterator, overload, Any, Type


class BaseToken(Protocol):
    """Base token protocol.

    This represents the static (unchangeable) parts of a token.
    """

    @property
    def text(self) -> str:
        """The text represented by this token."""
        pass

    @property
    def lower(self) -> str:
        """The lower case text representation."""
        pass

    @property
    def text_versions(self) -> list[str]:
        """The different versions of text (e.g normalised and lower)"""
        pass

    @property
    def is_upper(self) -> bool:
        """Whether the text is upper case."""
        pass

    @property
    def is_stop(self) -> bool:
        """Whether the token represents a stop token."""
        pass

    @property
    def char_index(self) -> int:
        """The character index of the start of this token"""
        pass

    @property
    def index(self) -> int:
        """The index (in terms of tokens) of this token in the document."""
        pass

    @property
    def text_with_ws(self) -> str:
        """The text with tailing whitespace (where applicable)."""
        pass

    @property
    def is_digit(self) -> bool:
        """Whether the token represents a digit."""


class MutableToken(Protocol):
    """The mutable part of a token.

    This protocol describes all the parts of a token that could be expected to
    change.
    """

    @property
    def base(self) -> BaseToken:
        """The base portion of the token."""
        pass

    @property
    def is_punctuation(self) -> bool:
        """Whether the token represents punctuation."""
        pass

    @is_punctuation.setter
    def is_punctuation(self, val: bool) -> None:
        """Whether the token represents punctuation."""
        pass

    @property
    def to_skip(self) -> bool:
        """Whether the token should be skipped."""
        pass

    @to_skip.setter
    def to_skip(self, new_val: bool) -> None:
        """Whether the token should be skipped."""
        pass

    @property
    def lemma(self) -> str:
        """The lemmatised version of the text."""
        pass

    @property
    def tag(self) -> Optional[str]:
        """Optional tag (e.g) for normalization."""
        pass

    @property
    def norm(self) -> str:
        """The normalised text."""
        pass

    @norm.setter
    def norm(self, value: str) -> None:
        """The normalised text."""
        pass


class BaseEntity(Protocol):
    """Base entity protocol.

    This describes the static (unchangeable) parts of an entity or
    sequence of tokens.
    """

    @property
    def start_index(self) -> int:
        """The index of the first token in the entity."""
        pass

    @property
    def end_index(self) -> int:
        """The index of the last token in the entity."""
        pass

    @property
    def start_char_index(self) -> int:
        """The character index of the first token."""
        pass

    @property
    def end_char_index(self) -> int:
        """The character index of the last token."""
        pass

    @property
    def label(self) -> int:
        """The label of the entity (NOTE: seems unused)."""
        pass

    @property
    def text(self) -> str:
        """The text of the entire entity."""
        pass

    def __iter__(self) -> Iterator[BaseToken]:
        pass

    def __len__(self) -> int:
        pass


class MutableEntity(Protocol):
    """The mutable part of an entity.

    This represent the changeable part of an entnity. That is, parts
    that should be changed by the various components.
    """

    @property
    def base(self) -> BaseEntity:
        """The base / static entity part."""
        pass

    @property
    def detected_name(self) -> str:
        """The detected name (if any) for this entity.

        This should be set by the NER component.
        """
        pass

    @detected_name.setter
    def detected_name(self, name: str) -> None:
        """The detected name (if any) for this entity.

        This should be set by the NER component.
        """
        pass

    def set_addon_data(self, path: str, val: Any) -> None:
        """Used to add arbitrary data to the entity.

        This is generally used by addons to keep track of their data.

        NB! The path used needs to be registered using the
        `register_addon_path` class method.

        Args:
            path (str): The data ID / path.
            val (Any): The value to be added.
        """
        pass

    def has_addon_data(self, path: str) -> bool:
        """Checks whether the addon data for a specific path has been set.

        Args:
            path (str): The path to check.

        Returns:
            bool: Whether the addon data had been set.
        """
        pass

    def get_addon_data(self, path: str) -> Any:
        """Get data added to the entity.

        See `add_data` for details.

        Args:
            path (str): The data ID / path.

        Returns:
            Any: The stored value.
        """
        pass

    def get_available_addon_paths(self) -> list[str]:
        """Gets the available addon data paths for this entity.

        This will only include paths that have values set.

        Returns:
            list[str]: List of available addon data paths.
        """
        pass

    @property
    def link_candidates(self) -> list[str]:
        """The candidates for the detected name (if any) for this entity.

        This should be set by the NER component.
        """
        pass

    @link_candidates.setter
    def link_candidates(self, candidates: list[str]) -> None:
        """The candidates for the detected name (if any) for this entity.

        This should be set by the NER component.
        """
        pass

    @property
    def context_similarity(self) -> float:
        """The context similarity of the lnked entity.

        This should be set by the linker component.
        """
        pass

    @context_similarity.setter
    def context_similarity(self, val: float) -> None:
        """The context similarity of the lnked entity.

        This should be set by the linker component.
        """
        pass

    @property
    def confidence(self) -> float:
        """The confidence for the lnked entity.

        NOTE: This seems to be unused!
        """
        pass

    @confidence.setter
    def confidence(self, val: float) -> None:
        """The confidence for the lnked entity.

        NOTE: This seems to be unused!
        """
        pass

    @property
    def cui(self) -> str:
        """The CUI of the lnked entity.

        This should be set by the linker component.
        """
        pass

    @cui.setter
    def cui(self, value: str) -> None:
        """The CUI of the lnked entity.

        This should be set by the linker component.
        """
        pass

    @property
    def id(self) -> int:
        """The ID of the entity within the document.

        This counts all the entities recognised, not just ones that
        were successfully linked.

        This should be set by the NER.
        """
        pass

    @id.setter
    def id(self, value: int) -> None:
        """The ID of the entity within the document.

        This counts all the entities recognised, not just ones that
        were successfully linked.

        This should be set by the NER.
        """
        pass

    @classmethod
    def register_addon_path(cls, path: str, def_val: Any = None,
                            force: bool = True) -> None:
        """Register a custom/arbitrary data path.

        This can be used to store arbitrary data along with the entity for
        use in an addon (e.g MetaCAT).

        PS: If using this, it is important to use paths namespaced to the
        component you're using in order to avoid conflicts.

        Args:
            path (str): The path to be used. Should be prefixed by component
                name (e.g `meta_cat_id` for an ID tied to the `meta_cat` addon)
            def_val (Any): Default value. Defaults to `None`.
            force (bool): Whether to forcefully add the value.
                Defaults to True.
        """
        pass

    def __iter__(self) -> Iterator[MutableToken]:
        pass

    def __len__(self) -> int:
        pass


class BaseDocument(Protocol):
    """The base document protocol.

    Represents the unchangeable parts of the whole document.
    """

    @property
    def text(self) -> str:
        """The document raw text."""
        pass

    @overload
    def __getitem__(self, index: int) -> BaseToken:
        pass

    @overload
    def __getitem__(self, index: slice) -> BaseEntity:
        pass

    def __iter__(self) -> Iterator[BaseToken]:
        pass

    def isupper(self) -> bool:
        """Whether the entire document is upper case."""
        pass


class MutableDocument(Protocol):
    """The mutable parts of the document.

    Represents parts of the document that can / should be changed
    by the various components.
    """

    @property
    def base(self) -> BaseDocument:
        """The base document."""
        pass

    @property
    def linked_ents(self) -> list[MutableEntity]:
        """The linked entities associated with the document.

        This should be set by the linker.
        """
        pass

    @property
    def ner_ents(self) -> list[MutableEntity]:
        """All entities recognised by NER.

        This should be set by the NER component.
        """
        pass

    def __iter__(self) -> Iterator[MutableToken]:
        pass

    @overload
    def __getitem__(self, index: int) -> MutableToken:
        pass

    @overload
    def __getitem__(self, index: slice) -> MutableEntity:
        pass

    def __len__(self) -> int:
        pass

    def get_tokens(self, start_index: int, end_index: int
                   ) -> list[MutableToken]:
        """Get the tokens that span the specified character indices.

        Args:
            start_index (int): The starting character index.
            end_index (int): The ending character index.

        Returns:
            list[MutableToken]:
                The list of tokens.
        """
        pass

    def set_addon_data(self, path: str, val: Any) -> None:
        """Used to add arbitrary data to the entity.

        This is generally used by addons to keep track of their data.

        NB! The path used needs to be registered using the
        `register_addon_path` class method.

        Args:
            path (str): The data ID / path.
            val (Any): The value to be added.
        """
        pass

    def has_addon_data(self, path: str) -> bool:
        """Checks whether the addon data for a specific path has been set.

        Args:
            path (str): The path to check.

        Returns:
            bool: Whether the addon data had been set.
        """
        pass

    def get_addon_data(self, path: str) -> Any:
        """Get data added to the entity.

        See `add_data` for details.

        Args:
            path (str): The data ID / path.

        Returns:
            Any: The stored value.
        """
        pass

    def get_available_addon_paths(self) -> list[str]:
        """Gets the available addon data paths for this document.

        This will only include paths that have values set.

        Returns:
            list[str]: List of available addon data paths.
        """
        pass

    @classmethod
    def register_addon_path(cls, path: str, def_val: Any = None,
                            force: bool = True) -> None:
        """Register a custom/arbitrary data path.

        This can be used to store arbitrary data along with the entity for
        use in an addon (e.g MetaCAT).

        PS: If using this, it is important to use paths namespaced to the
        component you're using in order to avoid conflicts.

        Args:
            path (str): The path to be used. Should be prefixed by component
                name (e.g `meta_cat_id` for an ID tied to the `meta_cat` addon)
            def_val (Any): Default value. Defaults to `None`.
            force (bool): Whether to forcefully add the value.
                Defaults to True.
        """
        pass


class UnregisteredDataPathException(ValueError):

    def __init__(self, cls: Type, path: str):
        super().__init__(
            f"Unregistered path '{path}' for class: {cls}")
        self.cls = cls
        self.path = path



from typing import Optional, Iterable, Iterator, Union, overload, Literal
import re

from medcat.tokenizing.tokens import MutableDocument
from medcat.tokenizing.tokenizers import BaseTokenizer
from medcat.config.config import Config, ComponentConfig
from medcat.vocab import Vocab
from medcat.cdb import CDB
from medcat.components.types import CoreComponentType, AbstractCoreComponent


CONTAINS_NUMBER = re.compile('[0-9]+')


class BasicSpellChecker:

    def __init__(self, cdb_vocab: dict[str, int], config: Config,
                 data_vocab: Optional[Vocab] = None):
        self.vocab = cdb_vocab
        self.config = config
        self.data_vocab = data_vocab

    def P(self, word: str) -> float:
        """Probability of `word`.

        Args:
            word (str): The word in question.

        Returns:
            float: The probability.
        """
    # use inverse of rank as proxy
    # returns 0 if the word isn't in the dictionary
        cnt = self.vocab.get(word, 0)
        if cnt != 0:
            return -1 / cnt
        else:
            return 0

    def __contains__(self, word):
        if word in self.vocab:
            return True
        elif self.data_vocab is not None and word in self.data_vocab:
            return False
        else:
            return False

    def fix(self, word: str) -> Optional[str]:
        """Most probable spelling correction for word.

        Args:
            word (str): The word.

        Returns:
            Optional[str]: Fixed word, or None if no fixes were applied.
        """
        fix = max(self.candidates(word), key=self.P)
        if fix != word:
            return fix
        else:
            return None

    def candidates(self, word: str) -> Iterable[str]:
        """Generate possible spelling corrections for word.

        Args:
            word (str): The word.

        Returns:
            Iterable[str]: The list of candidate words.
        """
        if self.config.general.spell_check_deep:
            # This will check a two letter edit distance
            return (self.known([word]) or
                    self.known(self.edits1(word)) or
                    self.known(self.edits2(word)) or
                    [word])
        else:
            # Will check only one letter edit distance
            return (self.known([word]) or
                    self.known(self.edits1(word)) or
                    [word])

    def known(self, words: Iterable[str]) -> set[str]:
        """The subset of `words` that appear in the dictionary of WORDS.

        Args:
            words (Iterable[str]): The words.

        Returns:
            set[str]: The set of candidates.
        """
        return set(w for w in words if w in self.vocab)

    def edits1(self, word: str) -> set[str]:
        """All edits that are one edit away from `word`.

        Args:
            word (str): The word.

        Returns:
            set[str]: The set of all edits
        """
        return self.raw_edits1(word, self.config.general.diacritics)

    @overload
    @classmethod
    def raw_edits1(cls, word: str, use_diacritics: bool = False,
                   return_ordered: Literal[False] = False) -> set[str]:
        pass

    @overload
    @classmethod
    def raw_edits1(cls, word: str, use_diacritics: bool = False,
                   return_ordered: Literal[True] = True) -> list[str]:
        pass

    @overload
    @classmethod
    def raw_edits1(cls, word: str, use_diacritics: bool = False,
                   return_ordered: bool = False) -> Union[set[str], list[str]]:
        pass

    @classmethod
    def raw_edits1(cls, word: str, use_diacritics: bool = False,
                   return_ordered: bool = False) -> Union[set[str], list[str]]:
        letters = 'abcdefghijklmnopqrstuvwxyz'

        if use_diacritics:
            letters += 'àáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ'

        splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
        deletes: list[str] = []
        transposes: list[str] = []
        replaces: list[str] = []
        inserts: list[str] = []
        for L, R in splits:
            if R:
                deletes.append(L + R[1:])
            if len(R) > 1:
                transposes.append(L + R[1] + R[0] + R[2:])
            if R:
                replaces.extend(L + c + R[1:] for c in letters)
            inserts.extend([L + c + R for c in letters])
        if not return_ordered:
            return set(deletes + transposes + replaces + inserts)
        else:
            return sorted(deletes + transposes + replaces + inserts)

    def edits2(self, word: str) -> Iterator[str]:
        """All edits that are two edits away from `word`.

        Args:
            word (str): The word to start from.

        Returns:
            Iterator[str]: All 2-away edits.
        """
        return self.raw_edits2(word, self.config.general.diacritics)

    @classmethod
    def raw_edits2(cls, word: str, use_diacritics: bool = False,
                   return_ordered: bool = False) -> Iterator[str]:
        return (
            e2 for e1 in cls.raw_edits1(word, use_diacritics, return_ordered)
            for e2 in cls.raw_edits1(e1, use_diacritics, return_ordered))

    def edits3(self, word):
        """All edits that are two edits away from `word`."""  # noqa
        # Do d3 edits
        raise ValueError("No implementation")


class TokenNormalizer(AbstractCoreComponent):
    """Will normalize all tokens in a spacy document.    """
    name = 'token_normalizer'

    # Override
    def __init__(self, nlp: BaseTokenizer, config: Config,
                 cdb_vocab: dict[str, int],
                 data_vocab: Optional[Vocab] = None):
        self.config = config
        self.spell_checker = BasicSpellChecker(cdb_vocab, config, data_vocab)
        self.nlp = nlp

    def get_type(self) -> CoreComponentType:
        return CoreComponentType.token_normalizing

    # Override
    def __call__(self, doc: MutableDocument):
        # avoid accessing all these in loop
        spell_check_limit = self.config.general.spell_check_len_limit
        min_len_normalizer = self.config.preprocessing.min_len_normalize
        do_not_normalize = self.config.preprocessing.do_not_normalize
        perform_spell_check = self.config.general.spell_check
        for token in doc:
            if len(token.base.lower) < min_len_normalizer:
                token.norm = token.base.lower
            elif (do_not_normalize and
                    token.tag is not None and
                    token.tag in do_not_normalize):
                token.norm = token.base.lower
            elif token.lemma == '-PRON-':
                token.norm = token.lemma
                token.to_skip = True
            else:
                token.norm = token.lemma.lower()

            if perform_spell_check:
                # Fix the token if necessary
                if (len(token.base.text) >= spell_check_limit and
                        not token.is_punctuation and self.spell_checker and
                        token.base.lower not in self.spell_checker and
                        not CONTAINS_NUMBER.search(token.base.lower)):
                    fix = self.spell_checker.fix(token.base.lower)
                    if fix is not None:
                        tmp = self.nlp(fix)[0]
                        if len(token.base.lower) < min_len_normalizer:
                            token.norm = tmp.base.lower
                        else:
                            token.norm = tmp.lemma.lower()
        return doc

    @classmethod
    def create_new_component(
            cls, cnf: ComponentConfig, tokenizer: BaseTokenizer,
            cdb: CDB, vocab: Vocab, model_load_path: Optional[str]
            ) -> 'TokenNormalizer':
        return cls(tokenizer, cdb.config, cdb.token_counts, vocab)



from typing import Any, Optional
from collections.abc import Container

from medcat.config.config import DirtiableBaseModel, ComponentConfig
from medcat.storage.serialisers import AvailableSerialisers

import logging


logger = logging.getLogger(__name__)


class General(DirtiableBaseModel):
    """The General part of the MetaCAT config"""
    device: str = 'cpu'
    """
    Device to used by the module to perform predicting/training.

    Reference:
        https://pytorch.org/docs/stable/tensor_attributes.html#torch.device
    """
    disable_component_lock: bool = False
    """ Whether to use the MetaCAT component lock.

    If set to False (the default), a component lock is used that forces
    usage only on one thread at a time.

    If set to True, the component lock is not used."""
    seed: int = 13
    """The seed for random number generation.

    NOTE: If used along RelCAT or additional NER, only one of the seeds
    will take effect
    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    description: str = "No description"
    """Should provide a basic description of this MetaCAT model"""
    category_name: Optional[str] = None
    """What category is this meta_cat model predicting/training.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    alternative_category_names: list[str] = []
    """List that stores the variations of possible category names
    Example: For Experiencer, the alternate name is Subject
    alternative_category_names: ['Experiencer','Subject']
    In the case that one specified in `category_name` parameter does not match
    the data, this ensures no error is raised and it is automatically mapped
    """
    category_value2id: dict[str, int] = {}
    """Map from category values to ID, if empty it will be autocalculated
    during training"""
    alternative_class_names: list[list[str]] = [[]]
    """List of lists that stores the variations of possible class names
    for each class mentioned in self.general.category_value2id.
    Example: For Presence task, the class names vary across NHS sites.
    To accommodate for this, alternative_class_names is populated as:
        [
         ["Hypothetical (N/A)","Hypothetical"],
         ["Not present (False)","False"],
         ["Present (True)","True"]
        ]
    Each sub list contains the possible variations of the given class.
    """
    vocab_size: int = -1
    """Will be set automatically if the tokenizer is provided during
    meta_cat init"""
    lowercase: bool = True
    """If true all input text will be lowercased"""
    cntx_left: int = 15
    """Number of tokens to take from the left of the concept"""
    cntx_right: int = 10
    """Number of tokens to take from the right of the concept"""
    replace_center: Optional[Any] = None
    """If set the center (concept) will be replaced with this string"""
    batch_size_eval: int = 5000
    """Number of annotations to be meta-annotated at once in eval"""
    tokenizer_name: str = 'bbpe'
    """
    Tokenizer name used with MetaCAT.

    Choose from:
        - 'bbpe': Byte Pair Encoding Tokenizer
        - 'bert-tokenizer': BERT Tokenizer

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    save_and_reuse_tokens: bool = False
    """This is a dangerous option, if not sure ALWAYS set to False. If set,
    it will try to share the pre-calculated context tokens between MetaCAT
    models when serving. It will ignore differences in tokenizer and context
    size, so you need to be sure that the models for which this is turned on
    have the same tokenizer and context size, during a deployment."""
    pipe_batch_size_in_chars: int = 20000000
    """How many characters are piped at once into the meta_cat class"""
    span_group: Optional[str] = None
    """If set, the spacy span group that the metacat model will assign
    annotations. Otherwise defaults to doc._.ents or doc.ents per the
    annotate_overlapping settings"""
    serialiser: AvailableSerialisers = AvailableSerialisers.dill
    """The serialiser to use when saving."""

    def get_applicable_category_name(
            self, available_names: Container[str]) -> Optional[str]:
        if self.category_name in available_names:
            return self.category_name
        matches = [cat for cat in self.alternative_category_names
                   if cat in available_names]
        if len(matches) > 0:
            logger.info(
                "The category name provided in the config - '%s' is not "
                "present in the data. However, the corresponding name - '%s' "
                "from the category_name_mapping has been found. Updating the "
                "category name...", self.category_name, *matches)
            self.category_name = matches[0]
            return self.category_name
        return None

    class Config:
        extra = 'allow'
        validate_assignment = True


class Model(DirtiableBaseModel):
    """The model part of the metaCAT config"""
    model_name: str = 'lstm'
    """
    Model to be used for training or predicting.

    Choose from:
        - 'bert'
        - 'lstm'

    Note:
        When changing the model, make sure to change the tokenizer accordingly.
        NB! For these changes to take effect, the pipe would need to
        be recreated.
    """
    model_variant: str = 'bert-base-uncased'
    """
    Applicable only when using BERT:

    Specifies the model variant to be used.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    model_freeze_layers: bool = True
    """
    Applicable only when using BERT:

    Determines the training approach for BERT.

    - If True: BERT layers are frozen and only the fully connected (FC)
                layer(s) on top are trained.
    - If False: Parameter-efficient fine-tuning will be applied using
                Low-Rank Adaptation (LoRA).

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    num_layers: int = 2
    """Number of layers in the model (both LSTM and BERT)

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    input_size: int = 300
    """
    Specifies the size of the embedding layer.

    Applicable only for LSTM model and ignored for BERT as BERT's embedding
    size is predefined.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    hidden_size: int = 300
    """Number of neurons in the hidden layer.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    dropout: float = 0.5
    """The dropout for the model.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    phase_number: int = 0
    """Indicates whether two phase learning is to be used for training.

    1: Phase 1 - Train model on undersampled data

    2: Phase 2 - Continue training on full data

    0: None - 2 phase learning is not performed

    Paper reference - https://ieeexplore.ieee.org/document/7533053"""
    category_undersample: str = ''
    """When using 2 phase learning, this category is used to undersample
    the data"""
    model_architecture_config: dict[str, bool] = {
        'fc2': True, 'fc3': False, 'lr_scheduler': True}
    """Specifies the architecture for BERT model.

    If fc2 is set to True, then the 2nd fully connected layer is used

    If fc2 is True and fc3 is set to True, then the 3rd fully connected
    layer is used

    If lr_scheduler is set to True, then the learning rate scheduler is
    used with the optimizer

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    num_directions: int = 2
    """Applicable only for LSTM:

    2 - bidirectional model, 1 - unidirectional

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    nclasses: int = 2
    """Number of classes that this model will output.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    padding_idx: int = -1
    """The padding index.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    emb_grad: bool = True
    """Applicable only for LSTM:

    If True, the embeddings will also be trained.

    NB! For these changes to take effect, the pipe would need to be recreated.
    """
    ignore_cpos: bool = False
    """If set to True center positions will be ignored when calculating
    representation"""

    class Config:
        extra = 'allow'
        validate_assignment = True
        protected_namespaces = ()


class Train(DirtiableBaseModel):
    """The train part of the metaCAT config"""
    batch_size: int = 100
    nepochs: int = 50
    lr: float = 0.001
    test_size: float = 0.1
    shuffle_data: bool = True
    """Used only during training, if set the dataset will be shuffled before
    train/test split"""
    class_weights: Optional[Any] = None
    compute_class_weights: bool = False
    """If true and class weights not provided, the class weights will be
    calculated based on the data"""
    score_average: str = 'weighted'
    """What to use for averaging F1/P/R across labels"""
    prerequisites: dict = {}
    cui_filter: Optional[Any] = None
    """If set only this CUIs will be used for training"""
    auto_save_model: bool = True
    """Should do model be saved during training for best results"""
    last_train_on: Optional[float] = None
    """When was the last training run"""
    metric: dict[str, str] = {'base': 'weighted avg', 'score': 'f1-score'}
    """What metric should be used for choosing the best model"""
    loss_funct: str = 'cross_entropy'
    """Loss function for the model.

    Choose from:
        - 'cross_entropy'
        - 'focal_loss'
    """
    gamma: int = 2
    """Focal Loss hyperparameter - determines importance the loss gives to
    hard-to-classify examples"""

    class Config:
        extra = 'allow'
        validate_assignment = True


class ConfigMetaCAT(ComponentConfig):
    """The MetaCAT part of the config"""
    comp_name: str = 'meta_cat'
    general: General = General()
    model: Model = Model()
    train: Train = Train()

    class Config:
        extra = 'allow'
        validate_assignment = True



from typing import Optional
import os
import argparse
import logging

from medcat.utils.legacy.conversion_all import Converter, logger as clogger
from medcat.storage.serialisers import AvailableSerialisers
from medcat.utils.legacy.helpers import logger as hlogger


def do_conversion(file_from: str, file_to: str,
                  save_format: AvailableSerialisers):
    """Perform the conversion from v1 to v2.

    Args:
        file_from (str): The v1 model pack path.
        file_to (str): The v2/target model pack path.
        save_format (AvailableSerialisers): The type of serialiser for v2.
    """
    converter = Converter(file_from, file_to, save_format)
    converter.convert()


def main(argv: Optional[list[str]] = None):
    parser = argparse.ArgumentParser()
    parser.add_argument('old_model', help='The path to the old model')
    parser.add_argument('new_model', help='The path to the new model')
    parser.add_argument('--format', help='The save format',
                        default=AvailableSerialisers.dill, required=False,
                        choices=list(AvailableSerialisers),
                        type=lambda s: AvailableSerialisers[s.lower()])
    parser.add_argument('--silent', '-s',
                        help='Make the operation silent (no console output)',
                        action='store_true')
    parser.add_argument('--verbose', '--debug',
                        help='Make the operation produce more debug output',
                        action='store_true')
    parser.add_argument('--new-folder', help='Create new folder if folder '
                        'does not exist', action='store_true')
    args = parser.parse_args(args=argv)
    if not args.silent:
        sh = logging.StreamHandler()
        clogger.addHandler(sh)
        hlogger.addHandler(sh)
    if args.verbose:
        clogger.setLevel(logging.DEBUG)
        hlogger.setLevel(logging.DEBUG)
    if not os.path.exists(args.new_model) and args.new_folder:
        print("Creating new folder:", args.new_model)
        os.mkdir(args.new_model)
    do_conversion(args.old_model, args.new_model, args.format)


if __name__ == "__main__":
    main()



import logging
from typing import Any, Optional
import os

import torch
from torch import nn
from transformers.models.llama import LlamaModel

from medcat.config.config_rel_cat import ConfigRelCAT
from medcat.components.addons.relation_extraction.llama.config import (
    RelExtrLlamaConfig)
from medcat.components.addons.relation_extraction.models import (
    RelExtrBaseModel)
from medcat.components.addons.relation_extraction.ml_utils import (
    create_dense_layers, get_annotation_schema_tag)


logger = logging.getLogger(__name__)


class RelExtrLlamaModel(RelExtrBaseModel):
    """ LlamaModel class for RelCAT
    """

    name = "llamamodel_relcat"

    def __init__(self, pretrained_model_name_or_path: str,
                 relcat_config: ConfigRelCAT,
                 model_config: RelExtrLlamaConfig):
        """Class to hold the Llama model + model_config

        Args:
            pretrained_model_name_or_path (str): path to load the model from,
                    this can be a HF model i.e: "bert-base-uncased",
                    if left empty, it is normally assumed that a model
                    is loaded from 'model.dat' using the RelCAT.load() method.
                    So if you are initializing/training a model from scratch
                    be sure to base it on some model.
            relcat_config (ConfigRelCAT): relcat config.
            model_config (Union[RelExtrBaseConfig | RelExtrLlamaConfig]):
                HF bert config for model.
        """

        super().__init__(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            relcat_config=relcat_config, model_config=model_config)

        self.relcat_config = relcat_config
        self.model_config = model_config

        self.hf_model = LlamaModel(config=model_config.hf_model_config)

        if pretrained_model_name_or_path != "":
            self.hf_model = LlamaModel.from_pretrained(
                pretrained_model_name_or_path, config=model_config,
                ignore_mismatched_sizes=True)

        for param in self.hf_model.parameters():
            param.requires_grad = False

        self.drop_out = nn.Dropout(self.relcat_config.model.dropout)

        # dense layers
        self.fc1, self.fc2, self.fc3 = create_dense_layers(self.relcat_config)

        # for pooled output
        self.llama_pooler = LlamaPooler(self.model_config.hidden_size)

        logger.info("RelCAT LlamaConfig: " + str(self.model_config))

    def output2logits(self, pooled_output: torch.Tensor,
                      sequence_output: torch.Tensor, input_ids: torch.Tensor,
                      e1_e2_start: torch.Tensor) -> torch.Tensor:
        """

        Args:
            pooled_output (torch.Tensor): embedding of the CLS token
            sequence_output (torch.Tensor): hidden states/embeddings for
                each token in the input text
            input_ids (torch.Tensor): input token ids.
            e1_e2_start (torch.Tensor): annotation tags token position

        Returns:
            torch.Tensor: classification probabilities for each token.
        """

        new_pooled_output = pooled_output

        if self.relcat_config.general.annotation_schema_tag_ids:
            annotation_schema_tag_ids_ = [
                self.relcat_config.general.annotation_schema_tag_ids[i:i + 2]
                for i in
                range(0, len(
                    self.relcat_config.general.annotation_schema_tag_ids), 2)]
            seq_tags_list: list[torch.Tensor] = []

            # for each pair of tags (e1,s1) and (e2,s2)
            for each_tags in annotation_schema_tag_ids_:
                seq_tags_list.append(get_annotation_schema_tag(
                    sequence_output, input_ids, each_tags))

            seq_tags = torch.stack(seq_tags_list, dim=0)

            if self.relcat_config.model.llama_use_pooled_output:
                new_pooled_output = torch.cat(
                    (pooled_output, *seq_tags), dim=1)
            else:
                new_pooled_output = torch.cat(
                    (seq_tags[0], seq_tags[1]), dim=1)
        else:
            e1e2_output = []
            temp_e1 = []
            temp_e2 = []

            for i, seq in enumerate(sequence_output):
                # e1e2 token sequences
                temp_e1.append(seq[e1_e2_start[i][0]])
                temp_e2.append(seq[e1_e2_start[i][1]])

            e1e2_output.append(torch.stack(temp_e1, dim=0))
            e1e2_output.append(torch.stack(temp_e2, dim=0))

            new_pooled_output = torch.cat((pooled_output, *e1e2_output), dim=1)

            del e1e2_output
            del temp_e2
            del temp_e1

        x = self.drop_out(new_pooled_output)
        x = self.fc1(x)
        x = self.drop_out(x)
        x = self.fc2(x)
        classification_logits = self.fc3(x)
        return classification_logits.to(self.relcat_config.general.device)

    def forward(
            self,
            input_ids: Optional[torch.Tensor] = None,
            attention_mask: Optional[torch.Tensor] = None,
            token_type_ids: Optional[torch.Tensor] = None,
            position_ids: Any = None,
            head_mask: Any = None,
            encoder_hidden_states: Any = None,
            encoder_attention_mask: Any = None,
            Q: Any = None,
            e1_e2_start: Any = None,
            pooled_output: Any = None) -> tuple[torch.Tensor, torch.Tensor]:

        if input_ids is not None:
            input_shape = input_ids.size()
        else:
            raise ValueError("You have to specify input_ids")

        if attention_mask is None:
            attention_mask = torch.ones(
                input_shape, device=self.relcat_config.general.device)
        if encoder_attention_mask is None:
            encoder_attention_mask = torch.ones(
                input_shape, device=self.relcat_config.general.device)
        if token_type_ids is None:
            token_type_ids = torch.zeros(
                input_shape, dtype=torch.long,
                device=self.relcat_config.general.device)

        input_ids = input_ids.to(self.relcat_config.general.device)
        attention_mask = attention_mask.to(self.relcat_config.general.device)
        encoder_attention_mask = encoder_attention_mask.to(
            self.relcat_config.general.device)

        # NOTE: the wrapping of the method means that mypy can't
        #       properly understand it
        self.hf_model = self.hf_model.to(
            self.relcat_config.general.device)  # type: ignore

        model_output = self.hf_model(input_ids=input_ids,
                                     attention_mask=attention_mask,
                                     output_hidden_states=True)

        # (batch_size, sequence_length, hidden_size)
        sequence_output = model_output.last_hidden_state

        if self.relcat_config.model.llama_use_pooled_output:
            pooled_output = self.llama_pooler(model_output)
            pooled_output = pooled_output.to(self.relcat_config.general.device)

        classification_logits = self.output2logits(
            pooled_output, sequence_output, input_ids, e1_e2_start)

        return model_output, classification_logits.to(
            self.relcat_config.general.device)

    @classmethod
    def load_specific(cls, pretrained_model_name_or_path: str,
                      relcat_config: ConfigRelCAT,
                      model_config: RelExtrLlamaConfig, **kwargs
                      ) -> "RelExtrLlamaModel":

        model = RelExtrLlamaModel(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            relcat_config=relcat_config, model_config=model_config)

        model_path = os.path.join(pretrained_model_name_or_path, "model.dat")

        if os.path.exists(model_path):
            model.load_state_dict(torch.load(
                model_path, map_location=relcat_config.general.device))
            logger.info("Loaded model from file: %s", model_path)
        elif pretrained_model_name_or_path:
            model.hf_model = LlamaModel.from_pretrained(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                config=model_config.hf_model_config,
                ignore_mismatched_sizes=True, **kwargs)
            logger.info("Loaded model from pretrained: %s",
                        pretrained_model_name_or_path)
        else:
            model.hf_model = LlamaModel.from_pretrained(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                config=model_config.hf_model_config,
                ignore_mismatched_sizes=True, **kwargs)
            logger.info("Loaded model from pretrained: %s",
                        cls.pretrained_model_name_or_path)

        return model


class LlamaPooler(nn.Module):
    """An attempt to copy the BERT pooling technique for an increase in
    performance.
    """

    def __init__(self, hidden_size: int):
        """ Initialises the pooler with a linear layer of size:
                self.model_config.hidden_size x self.model_config.hidden_size

        Args:
            hidden_size (int): size of tensor
        """
        super().__init__()
        self.dense = nn.Linear(hidden_size, hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token. (original BERT)
        # We can do the same here but the [CLS] token equivalent is not the
        # same as for bert as there is not much learning contained in it.
        # e.g: first_token_tensor = hidden_states[:, 0] # original
        # so instead we pool across all the tokens from the last hidden layer.

        pooled_output, _ = torch.max(hidden_states[-1], dim=1)
        pooled_output = self.dense(pooled_output)
        pooled_output = self.activation(pooled_output)

        return pooled_output
